# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

try:
    import numpy  # noqa: F401
except ImportError:
    pass
else:
    del numpy
    from .array_data_model import ArrayDataModel  # noqa: F401

from .data_accessors import (  # noqa: F401
    AbstractDataAccessor, AttributeDataAccessor, ConstantDataAccessor,
    IndexDataAccessor, KeyDataAccessor
)
from .row_table_data_model import RowTableDataModel  # noqa: F401



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
#
# Author: Enthought, Inc.
# Description: <Enthought pyface package component>
""" Abstract base class for all application actions. """

import platform


from traits.api import Instance, Property, cached_property


from pyface.action.listening_action import ListeningAction

IS_WINDOWS = platform.system() == "Windows"


class GUIApplicationAction(ListeningAction):
    """ Abstract base class for GUI Application actions. """

    # 'ListeningAction' interface --------------------------------------------

    object = Property(observe="application")

    # 'WindowAction' interface -----------------------------------------------

    #: The application that the action is associated with.
    application = Instance("pyface.gui_application.GUIApplication")

    # ------------------------------------------------------------------------
    # Protected interface.
    # ------------------------------------------------------------------------

    def _get_object(self):
        return self.application

    def destroy(self):
        # Disconnect listeners to application and dependent properties.
        self.application = None
        super(GUIApplicationAction, self).destroy()


class ActiveWindowAction(GUIApplicationAction):
    """ Abstract base class for application active window actions. """

    # 'ListeningAction' interface --------------------------------------------

    object = Property(observe="application.active_window")

    # ------------------------------------------------------------------------
    # Protected interface.
    # ------------------------------------------------------------------------

    @cached_property
    def _get_object(self):
        if self.application is not None:
            return self.application.active_window


class CreateWindowAction(GUIApplicationAction):
    """ A standard 'New Window' menu action. """

    name = "New Window"
    accelerator = "Ctrl+N"

    def perform(self, event=None):
        window = self.application.create_window()
        self.application.add_window(window)


class ExitAction(GUIApplicationAction):
    """ A standard 'Quit' or 'Exit' menu action. """

    accelerator = "Alt+F4" if IS_WINDOWS else "Ctrl+Q"
    method = "exit"

    def _name_default(self):
        return ("Exit " if IS_WINDOWS else "Quit ") + self.application.name


class AboutAction(GUIApplicationAction):
    """ A standard 'About' dialog menu action. """

    method = "do_about"

    def _name_default(self):
        return "About " + self.application.name


class CloseActiveWindowAction(ActiveWindowAction):
    """ A standard 'Close window' menu action at the application level.

    This method closes the active window of the application.
    """

    name = "Close Window"
    accelerator = "Ctrl+W"
    method = "close"



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
import unittest

from traits.testing.api import UnittestTools
from pyface.tasks.api import Task
from ..task_window import TaskWindow


def _task_window_with_named_tasks(*names, **kwargs):
    tasks = [Task(name=name) for name in names]

    first_active = kwargs.pop("first_active", False)
    if first_active:
        kwargs["active_task"] = tasks[0]

    task = TaskWindow(tasks=tasks, **kwargs)
    return task


class TestTaskWindow(unittest.TestCase, UnittestTools):
    def test_title_default(self):
        task_window = TaskWindow()

        # default is empty
        self.assertEqual(task_window.title, "")

    def test_title_no_active_task(self):
        task_window = _task_window_with_named_tasks("Test Task", "Test Task 2")

        # should be empty
        self.assertEqual(task_window.title, "")

    def test_title_activate_task(self):
        task_window = _task_window_with_named_tasks("Test Task")
        task = task_window.tasks[0]

        # activate task
        with self.assertTraitChanges(task_window, "title", count=1):
            task_window.active_task = task
        self.assertEqual(task_window.title, "Test Task")

    def test_title_change_active_task_name(self):
        task_window = _task_window_with_named_tasks(
            "Test Task", first_active=True
        )
        task_1 = task_window.tasks[0]

        # change task name
        with self.assertTraitChanges(task_window, "title", count=1):
            task_1.name = "Changed Name"
        self.assertEqual(task_window.title, "Changed Name")

    def test_title_change_active_task(self):
        task_window = _task_window_with_named_tasks(
            "Test Task 1", "Test Task 2", first_active=True
        )
        task = task_window.tasks[1]

        # change active task
        with self.assertTraitChanges(task_window, "title", count=1):
            task_window.active_task = task
        self.assertEqual(task_window.title, "Test Task 2")

    def test_title_change_deactivate_task(self):
        task_window = _task_window_with_named_tasks(
            "Test Task 1", first_active=True
        )

        # change active task
        with self.assertTraitChanges(task_window, "title", count=1):
            task_window.active_task = None
        self.assertEqual(task_window.title, "")

    def test_set_title_no_tasks(self):
        task_window = _task_window_with_named_tasks()

        # set window title
        with self.assertTraitChanges(task_window, "title", count=1):
            task_window.title = "Window title"
        self.assertEqual(task_window.title, "Window title")

    def test_set_title_change_title(self):
        task_window = _task_window_with_named_tasks(title="Window Title")

        # set window title
        with self.assertTraitChanges(task_window, "title", count=1):
            task_window.title = "New Window title"
        self.assertEqual(task_window.title, "New Window title")

    def test_set_title_no_active_task(self):
        task_window = _task_window_with_named_tasks("Test Task")

        # set window title
        with self.assertTraitChanges(task_window, "title", count=1):
            task_window.title = "Window title"
        self.assertEqual(task_window.title, "Window title")

    def test_set_title_active_task(self):
        task_window = _task_window_with_named_tasks(
            "Test Task", first_active=True
        )

        # set window title
        with self.assertTraitChanges(task_window, "title", count=1):
            task_window.title = "Window title"
        self.assertEqual(task_window.title, "Window title")

    def test_set_title_activate_task(self):
        task_window = _task_window_with_named_tasks(
            "Test Task", title="Window title"
        )
        task = task_window.tasks[0]

        # change activate task (trait fires, no window title change)
        with self.assertTraitChanges(task_window, "title", count=1):
            task_window.active_task = task
        self.assertEqual(task_window.title, "Window title")

    def test_set_title_change_active_task_name(self):
        task_window = _task_window_with_named_tasks(
            "Test Task", title="Window title", first_active=True
        )
        task = task_window.tasks[0]

        # change task name (trait fires, no window title change)
        with self.assertTraitChanges(task_window, "title", count=1):
            task.name = "Changed Name"
        self.assertEqual(task_window.title, "Window title")

    def test_set_title_change_active_task(self):
        task_window = _task_window_with_named_tasks(
            "Test Task", "Test Task 2", title="Window title", active_first=True
        )
        task = task_window.tasks[1]

        # change task name (trait fires, no window title change)
        with self.assertTraitChanges(task_window, "title", count=1):
            task_window.active_task = task
        self.assertEqual(task_window.title, "Window title")

    def test_reset_title_active_task(self):
        task_window = _task_window_with_named_tasks(
            "Test Task", title="Window title", first_active=True
        )

        # reset window title
        with self.assertTraitChanges(task_window, "title", count=1):
            task_window.title = ""
        self.assertEqual(task_window.title, "Test Task")

    def test_reset_title(self):
        task_window = _task_window_with_named_tasks(
            "Test Task", title="Window title"
        )

        # set window title
        with self.assertTraitChanges(task_window, "title", count=1):
            task_window.title = ""
        self.assertEqual(task_window.title, "")



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

from traits.trait_list_object import TraitList
from traits.testing.api import UnittestTools
from traits.testing.optional_dependencies import numpy as np, requires_numpy

from pyface.data_view.abstract_data_model import DataViewSetError
from pyface.data_view.abstract_value_type import AbstractValueType
from pyface.data_view.value_types.api import (
    FloatValue, IntValue, TextValue, no_value
)
from pyface.data_view.data_models.data_accessors import (
    AttributeDataAccessor, IndexDataAccessor, KeyDataAccessor
)
from pyface.data_view.data_models.row_table_data_model import RowTableDataModel


class DataItem:

    def __init__(self, a, b, c):
        self.a = a
        self.b = b
        self.c = c


class TestRowTableDataModel(UnittestTools, unittest.TestCase):

    def setUp(self):
        super().setUp()
        self.data = [
            DataItem(a=i, b=10*i, c=str(i)) for i in range(10)
        ]
        self.model = RowTableDataModel(
            data=self.data,
            row_header_data=AttributeDataAccessor(
                attr='a',
                value_type=IntValue(),
            ),
            column_data=[
                AttributeDataAccessor(
                    attr='b',
                    value_type=IntValue(),
                ),
                AttributeDataAccessor(
                    attr='c',
                    value_type=TextValue(),
                )
            ]
        )
        self.values_changed_event = None
        self.structure_changed_event = None
        self.model.observe(self.model_values_changed, 'values_changed')
        self.model.observe(self.model_structure_changed, 'structure_changed')

    def tearDown(self):
        self.model.observe(
            self.model_values_changed, 'values_changed', remove=True)
        self.model.observe(
            self.model_structure_changed, 'structure_changed', remove=True)
        self.values_changed_event = None
        self.structure_changed_event = None
        super().tearDown()

    def model_values_changed(self, event):
        self.values_changed_event = event

    def model_structure_changed(self, event):
        self.structure_changed_event = event

    def test_no_data(self):
        model = RowTableDataModel()
        self.assertEqual(model.get_column_count(), 0)
        self.assertTrue(model.can_have_children(()))
        self.assertEqual(model.get_row_count(()), 0)

    def test_get_column_count(self):
        result = self.model.get_column_count()
        self.assertEqual(result, 2)

    def test_can_have_children(self):
        for row in self.model.iter_rows():
            with self.subTest(row=row):
                result = self.model.can_have_children(row)
                if len(row) == 0:
                    self.assertEqual(result, True)
                else:
                    self.assertEqual(result, False)

    def test_get_row_count(self):
        for row in self.model.iter_rows():
            with self.subTest(row=row):
                result = self.model.get_row_count(row)
                if len(row) == 0:
                    self.assertEqual(result, 10)
                else:
                    self.assertEqual(result, 0)

    def test_get_value(self):
        for row, column in self.model.iter_items():
            with self.subTest(row=row, column=column):
                result = self.model.get_value(row, column)
                if len(row) == 0 and len(column) == 0:
                    self.assertEqual(result, 'A')
                elif len(row) == 0:
                    attr = self.model.column_data[column[0]].attr
                    self.assertEqual(result, attr.title())
                elif len(column) == 0:
                    self.assertEqual(result, row[0])
                else:
                    attr = self.model.column_data[column[0]].attr
                    self.assertEqual(
                        result,
                        getattr(self.data[row[0]], attr)
                    )

    def test_set_value(self):
        for row, column in self.model.iter_items():
            with self.subTest(row=row, column=column):
                if len(row) == 0 and len(column) == 0:
                    with self.assertRaises(DataViewSetError):
                        self.model.set_value(row, column, 0)
                elif len(row) == 0:
                    with self.assertRaises(DataViewSetError):
                        self.model.set_value(row, column, 0)
                elif len(column) == 0:
                    value = 6.0 * row[0]
                    with self.assertTraitChanges(self.model, "values_changed"):
                        self.model.set_value(row, column, value)
                    self.assertEqual(self.data[row[0]].a, value)
                    self.assertEqual(
                        self.values_changed_event.new,
                        (row, column, row, column)
                    )
                else:
                    value = 6.0 * row[-1] + 2 * column[0]
                    with self.assertTraitChanges(self.model, "values_changed"):
                        self.model.set_value(row, column, value)
                    attr = self.model.column_data[column[0]].attr
                    self.assertEqual(
                        getattr(self.data[row[0]], attr),
                        value,
                    )
                    self.assertEqual(
                        self.values_changed_event.new,
                        (row, column, row, column)
                    )

    def test_get_value_type(self):
        for row, column in self.model.iter_items():
            with self.subTest(row=row, column=column):
                result = self.model.get_value_type(row, column)
                if len(row) == 0 and len(column) == 0:
                    self.assertIsInstance(result, AbstractValueType)
                    self.assertIs(
                        result,
                        self.model.row_header_data.title_type,
                    )
                elif len(row) == 0:
                    self.assertIsInstance(result, AbstractValueType)
                    self.assertIs(
                        result,
                        self.model.column_data[column[0]].title_type,
                    )
                elif len(column) == 0:
                    self.assertIsInstance(result, AbstractValueType)
                    self.assertIs(
                        result,
                        self.model.row_header_data.value_type,
                    )
                else:
                    self.assertIsInstance(result, AbstractValueType)
                    self.assertIs(
                        result,
                        self.model.column_data[column[0]].value_type,
                    )

    def test_data_updated(self):
        with self.assertTraitChanges(self.model, "structure_changed"):
            self.model.data = [
                DataItem(a=i+1, b=20*(i+1), c=str(i)) for i in range(10)
            ]
        self.assertTrue(self.structure_changed_event.new)

    def test_data_items_updated_item_added(self):
        self.model.data = TraitList([
            DataItem(a=i, b=10*i, c=str(i)) for i in range(10)
        ])
        with self.assertTraitChanges(self.model, "structure_changed"):
            self.model.data += [DataItem(a=100, b=200, c="a string")]
        self.assertTrue(self.structure_changed_event.new)

    def test_data_items_updated_item_replaced(self):
        self.model.data = TraitList([
            DataItem(a=i, b=10*i, c=str(i)) for i in range(10)
        ])
        with self.assertTraitChanges(self.model, "values_changed"):
            self.model.data[1] = DataItem(a=100, b=200, c="a string")
        self.assertEqual(self.values_changed_event.new, ((1,), (), (1,), ()))

    def test_data_items_updated_item_replaced_negative(self):
        self.model.data = TraitList([
            DataItem(a=i, b=10*i, c=str(i)) for i in range(10)
        ])
        with self.assertTraitChanges(self.model, "values_changed"):
            self.model.data[-2] = DataItem(a=100, b=200, c="a string")
        self.assertEqual(self.values_changed_event.new, ((8,), (), (8,), ()))

    def test_data_items_updated_items_replaced(self):
        self.model.data = TraitList([
            DataItem(a=i, b=10*i, c=str(i)) for i in range(10)
        ])
        with self.assertTraitChanges(self.model, "values_changed"):
            self.model.data[1:3] = [
                DataItem(a=100, b=200, c="a string"),
                DataItem(a=200, b=300, c="another string"),
            ]
        self.assertEqual(self.values_changed_event.new, ((1,), (), (2,), ()))

    def test_data_items_updated_slice_replaced(self):
        self.model.data = TraitList([
            DataItem(a=i, b=10*i, c=str(i)) for i in range(10)
        ])
        with self.assertTraitChanges(self.model, "values_changed"):
            self.model.data[1:4:2] = [
                DataItem(a=100, b=200, c="a string"),
                DataItem(a=200, b=300, c="another string"),
            ]
        self.assertEqual(self.values_changed_event.new, ((1,), (), (3,), ()))

    def test_data_items_updated_reverse_slice_replaced(self):
        self.model.data = TraitList([
            DataItem(a=i, b=10*i, c=str(i)) for i in range(10)
        ])
        with self.assertTraitChanges(self.model, "values_changed"):
            self.model.data[3:1:-1] = [
                DataItem(a=100, b=200, c="a string"),
                DataItem(a=200, b=300, c="another string"),
            ]
        self.assertEqual(self.values_changed_event.new, ((2,), (), (3,), ()))

    def test_row_header_data_updated(self):
        with self.assertTraitChanges(self.model, "values_changed"):
            self.model.row_header_data = AttributeDataAccessor(attr='b')
        self.assertEqual(
            self.values_changed_event.new,
            ((), (), (), ())
        )

    def test_row_header_data_values_updated(self):
        with self.assertTraitChanges(self.model, "values_changed"):
            self.model.row_header_data.updated = (self.model.row_header_data, 'value')
        self.assertEqual(
            self.values_changed_event.new,
            ((0,), (), (9,), ())
        )

    def test_row_header_data_title_updated(self):
        with self.assertTraitChanges(self.model, "values_changed"):
            self.model.row_header_data.updated = (self.model.row_header_data, 'title')
        self.assertEqual(
            self.values_changed_event.new,
            ((), (), (), ())
        )

    def test_no_data_row_header_data_update(self):
        model = RowTableDataModel(
            row_header_data=AttributeDataAccessor(
                attr='a',
                value_type=IntValue(),
            ),
            column_data=[
                AttributeDataAccessor(
                    attr='b',
                    value_type=IntValue(),
                ),
                AttributeDataAccessor(
                    attr='c',
                    value_type=TextValue(),
                )
            ]
        )

        # check that updating accessors is safe with empty data
        with self.assertTraitDoesNotChange(model, 'values_changed'):
            model.row_header_data.attr = 'b'

    def test_column_data_updated(self):
        with self.assertTraitChanges(self.model, "structure_changed"):
            self.model.column_data = [
                AttributeDataAccessor(
                    attr='c',
                    value_type=TextValue(),
                ),
                AttributeDataAccessor(
                    attr='b',
                    value_type=IntValue(),
                ),
            ]
        self.assertTrue(self.structure_changed_event.new)

    def test_column_data_items_updated(self):
        with self.assertTraitChanges(self.model, "structure_changed"):
            self.model.column_data.pop()
        self.assertTrue(self.structure_changed_event.new)

    def test_column_data_value_updated(self):
        with self.assertTraitChanges(self.model, "values_changed"):
            self.model.column_data[0].updated = (self.model.column_data[0], 'value')
        self.assertEqual(
            self.values_changed_event.new,
            ((0,), (0,), (9,), (0,))
        )

    def test_no_data_column_data_update(self):
        model = RowTableDataModel(
            row_header_data=AttributeDataAccessor(
                attr='a',
                value_type=IntValue(),
            ),
            column_data=[
                AttributeDataAccessor(
                    attr='b',
                    value_type=IntValue(),
                ),
                AttributeDataAccessor(
                    attr='c',
                    value_type=TextValue(),
                )
            ]
        )

        with self.assertTraitDoesNotChange(model, 'values_changed'):
            model.column_data[0].attr = 'a'

    def test_column_data_title_updated(self):
        with self.assertTraitChanges(self.model, "values_changed"):
            self.model.column_data[0].updated = (self.model.column_data[0], 'title')
        self.assertEqual(
            self.values_changed_event.new,
            ((), (0,), (), (0,))
        )

    def test_list_tuple_data(self):
        data = [
            (i, 10*i, str(i)) for i in range(10)
        ]
        model = RowTableDataModel(
            data=data,
            row_header_data=IndexDataAccessor(
                index=0,
                value_type=IntValue(),
            ),
            column_data=[
                IndexDataAccessor(
                    index=1,
                    value_type=IntValue(),
                ),
                IndexDataAccessor(
                    index=2,
                    value_type=TextValue(),
                )
            ]
        )

        for row, column in model.iter_items():
            with self.subTest(row=row, column=column):
                result = model.get_value(row, column)
                if len(row) == 0 and len(column) == 0:
                    self.assertEqual(result, '0')
                elif len(row) == 0:
                    index = model.column_data[column[0]].index
                    self.assertEqual(result, str(index))
                elif len(column) == 0:
                    self.assertEqual(result, row[0])
                else:
                    index = model.column_data[column[0]].index
                    self.assertEqual(
                        result,
                        data[row[0]][index]
                    )

    def test_list_dict_data(self):
        data = [
            {'a': i, 'b': 10*i, 'c': str(i)} for i in range(10)
        ]
        model = RowTableDataModel(
            data=data,
            row_header_data=KeyDataAccessor(
                key='a',
                value_type=IntValue(),
            ),
            column_data=[
                KeyDataAccessor(
                    key='b',
                    value_type=IntValue(),
                ),
                KeyDataAccessor(
                    key='c',
                    value_type=TextValue(),
                )
            ]
        )

        for row, column in model.iter_items():
            with self.subTest(row=row, column=column):
                result = model.get_value(row, column)
                if len(row) == 0 and len(column) == 0:
                    self.assertEqual(result, 'A')
                elif len(row) == 0:
                    key = model.column_data[column[0]].key
                    self.assertEqual(result, str(key).title())
                elif len(column) == 0:
                    self.assertEqual(result, data[row[0]]['a'])
                else:
                    key = model.column_data[column[0]].key
                    self.assertEqual(
                        result,
                        data[row[0]][key]
                    )



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest


from pyface.action.api import (
    Action,
    ActionItem,
    ActionManager,
    Group,
    MenuManager,
    MenuBarManager,
)
from pyface.tasks.action.api import (
    GroupSchema,
    MenuSchema,
    MenuBarSchema,
    SchemaAddition,
)
from pyface.tasks.action.task_action_manager_builder import (
    TaskActionManagerBuilder,
)
from pyface.tasks.api import Task


class ActionManagerBuilderTestCase(unittest.TestCase):

    # 'TestCase' protocol -------------------------------------------------#

    def setUp(self):
        """ Create some dummy actions to use while testing.
        """
        for i in range(1, 7):
            action_id = "action%i" % i
            setattr(
                self, action_id, Action(id=action_id, name="Action %i" % i)
            )

    # 'ActionManagerBuilderTestCase' protocol -----------------------------#

    def assertActionElementsEqual(self, first, second):
        """ Checks that two action managers are (logically) equivalent.
        """
        children1 = children2 = []
        self.assertEqual(type(first), type(second))
        self.assertEqual(first.id, second.id)

        if isinstance(first, ActionItem):
            self.assertEqual(first.action.name, second.action.name)

        elif isinstance(first, ActionManager):
            if not isinstance(first, MenuBarManager):
                self.assertEqual(first.name, second.name)
            children1, children2 = first.groups, second.groups

        elif isinstance(first, Group):
            self.assertEqual(first.separator, second.separator)
            children1, children2 = first.items, second.items

        self.assertEqual(len(children1), len(children2))
        for i in range(len(children1)):
            self.assertActionElementsEqual(children1[i], children2[i])

    # Tests ----------------------------------------------------------------

    def test_simple_menu_bar(self):
        """ Does constructing a simple menu with no additions work?
        """
        schema = MenuBarSchema(
            MenuSchema(self.action1, self.action2, id="File", name="&File"),
            MenuSchema(self.action3, self.action4, id="Edit", name="&Edit"),
        )
        builder = TaskActionManagerBuilder(task=Task(menu_bar=schema))
        actual = builder.create_menu_bar_manager()
        desired = MenuBarManager(
            MenuManager(self.action1, self.action2, id="File", name="&File"),
            MenuManager(self.action3, self.action4, id="Edit", name="&Edit"),
            id="MenuBar",
        )
        self.assertActionElementsEqual(actual, desired)

    # Tests about schema additions -----------------------------------------

    def test_additions_menu_bar(self):
        """ Does constructing a menu with a few additions work?
        """
        schema = MenuBarSchema(
            MenuSchema(
                GroupSchema(self.action1, self.action2, id="FileGroup"),
                id="File",
            )
        )
        extras = [
            SchemaAddition(
                factory=lambda: self.action3,
                before="action1",
                path="MenuBar/File/FileGroup",
            ),
            SchemaAddition(
                factory=lambda: self.action4,
                before="action1",
                path="MenuBar/File/FileGroup",
            ),
            SchemaAddition(
                factory=lambda: self.action5, path="MenuBar/File/FileGroup"
            ),
        ]
        builder = TaskActionManagerBuilder(
            task=Task(menu_bar=schema, extra_actions=extras)
        )
        actual = builder.create_menu_bar_manager()
        desired = MenuBarManager(
            MenuManager(
                Group(
                    self.action3,
                    self.action4,
                    self.action1,
                    self.action2,
                    self.action5,
                    id="FileGroup",
                ),
                id="File",
            ),
            id="MenuBar",
        )
        self.assertActionElementsEqual(actual, desired)

    def test_extra_menu(self):
        """ Test contributing a whole new menu to the menu bar. """

        # Initial menu.
        schema = MenuBarSchema(
            MenuSchema(
                GroupSchema(self.action1, id="FileGroup"), id="FileMenu"
            )
        )

        # Contributed menu.
        extra_menu = MenuSchema(
            GroupSchema(self.action2, id="BarGroup"), id="DummyActionsMenu"
        )

        extra_actions = [
            SchemaAddition(
                path="MenuBar",
                factory=lambda: extra_menu,
                id="DummyActionsSMenu",
            )
        ]

        # Build the final menu.
        builder = TaskActionManagerBuilder(
            task=Task(menu_bar=schema, extra_actions=extra_actions)
        )
        actual = builder.create_menu_bar_manager()

        desired = MenuBarManager(
            MenuManager(Group(self.action1, id="FileGroup"), id="FileMenu"),
            MenuManager(
                Group(self.action2, id="BarGroup"), id="DummyActionsMenu"
            ),
            id="MenuBar",
        )

        self.assertActionElementsEqual(actual, desired)

    # Tests about merging schemas -----------------------------------------#

    def test_merging_redundant_items(self):
        """ Menus and groups with matching path are merged together. """

        # Initial menu.
        schema = MenuBarSchema(
            MenuSchema(
                GroupSchema(self.action1, id="FileGroup"),
                name="File menu number one",
                id="FileMenu",
            )
        )

        # Contributed menus.
        extra_menu = MenuSchema(
            GroupSchema(self.action2, id="FileGroup"),
            name="File menu number two",
            id="FileMenu",
        )

        extra_actions = [
            SchemaAddition(
                path="MenuBar",
                factory=lambda: extra_menu,
                id="DummyActionsSMenu",
            )
        ]

        # Build the final menu.
        builder = TaskActionManagerBuilder(
            task=Task(menu_bar=schema, extra_actions=extra_actions)
        )
        actual = builder.create_menu_bar_manager()

        # Note that we expect the name of the menu to be inherited from
        # the menu in the menu bar schema that is defined first.
        desired = MenuBarManager(
            MenuManager(
                Group(self.action1, self.action2, id="FileGroup"),
                name="File menu number one",
                id="FileMenu",
            ),
            id="MenuBar",
        )
        self.assertActionElementsEqual(actual, desired)

    def test_unwanted_merge(self):
        """ Test that we don't have automatic merges due to forgetting to set
        a schema ID. """

        # Initial menu.
        schema = MenuBarSchema(
            MenuSchema(
                GroupSchema(self.action1, id="FileGroup"), name="File 1"
            )
        )

        # Contributed menus.
        extra_menu = MenuSchema(
            GroupSchema(self.action2, id="FileGroup"), name="File 2"
        )

        extra_actions = [
            SchemaAddition(
                path="MenuBar",
                factory=lambda: extra_menu,
                id="DummyActionsSMenu",
            )
        ]

        # Build the final menu.
        builder = TaskActionManagerBuilder(
            task=Task(menu_bar=schema, extra_actions=extra_actions)
        )
        actual = builder.create_menu_bar_manager()

        # Note that we expect the name of the menu to be inherited from
        # the menu in the menu bar schema that is defined first.
        desired = MenuBarManager(
            MenuManager(
                Group(self.action1, id="FileGroup"),
                name="File 1",
                id="MenuSchema_1",
            ),
            MenuManager(
                Group(self.action2, id="FileGroup"),
                name="File 2",
                id="MenuSchema_2",
            ),
            id="MenuBar",
        )
        self.assertActionElementsEqual(actual, desired)

    def test_merging_items_with_same_id_but_different_class(self):
        """ Schemas with the same path but different types (menus, groups)
        are not merged together.

        Having a group and a menu with the same path is of course bad practice,
        but we need a predictable outcome.

        """

        # Initial menu.
        schema = MenuBarSchema(
            MenuSchema(
                GroupSchema(self.action1, id="FileGroup"), id="FileSchema"
            )
        )

        # Contributed menus.
        extra_group = GroupSchema(self.action2, id="FileSchema")

        extra_actions = [
            SchemaAddition(
                path="MenuBar",
                factory=(lambda: extra_group),
                id="DummyActionsSMenu",
            )
        ]

        # Build the final menu.
        builder = TaskActionManagerBuilder(
            task=Task(menu_bar=schema, extra_actions=extra_actions)
        )
        actual = builder.create_menu_bar_manager()

        desired = MenuBarManager(
            MenuManager(Group(self.action1, id="FileGroup"), id="FileSchema"),
            Group(self.action2, id="FileSchema"),
            id="MenuBar",
        )
        self.assertActionElementsEqual(actual, desired)

    def test_merging_redundant_items_that_are_not_schemas(self):
        """ Items that are not schemas cannot be merged, but we should
        not crash, either. """

        # Initial menu.
        schema = MenuBarSchema(
            # This menu is not a schema...
            MenuManager(Group(self.action1, id="FileGroup"), id="FileMenu")
        )

        # Contributed menus.
        extra_menu = MenuSchema(
            GroupSchema(self.action2, id="FileGroup"), id="FileMenu"
        )

        extra_actions = [
            SchemaAddition(
                path="MenuBar",
                factory=lambda: extra_menu,
                id="DummyActionsSMenu",
            )
        ]

        # Build the final menu.
        builder = TaskActionManagerBuilder(
            task=Task(menu_bar=schema, extra_actions=extra_actions)
        )
        actual = builder.create_menu_bar_manager()

        desired = MenuBarManager(
            MenuManager(Group(self.action1, id="FileGroup"), id="FileMenu"),
            MenuManager(Group(self.action2, id="FileGroup"), id="FileMenu"),
            id="MenuBar",
        )
        self.assertActionElementsEqual(actual, desired)

    # Tests about ordering -------------------------------------------------

    def test_absolute_ordering(self):
        """ Does specifying absolute_position work?
        """
        schema = MenuBarSchema(
            MenuSchema(
                GroupSchema(self.action1, self.action2, id="FileGroup"),
                id="File",
            )
        )
        extras = [
            SchemaAddition(
                factory=lambda: self.action3,
                absolute_position="last",
                path="MenuBar/File/FileGroup",
            ),
            SchemaAddition(
                factory=lambda: self.action4,
                absolute_position="first",
                path="MenuBar/File/FileGroup",
            ),
            SchemaAddition(
                factory=lambda: self.action5,
                absolute_position="first",
                path="MenuBar/File/FileGroup",
            ),
        ]
        builder = TaskActionManagerBuilder(
            task=Task(menu_bar=schema, extra_actions=extras)
        )
        actual = builder.create_menu_bar_manager()
        desired = MenuBarManager(
            MenuManager(
                Group(
                    self.action4,
                    self.action5,
                    self.action1,
                    self.action2,
                    self.action3,
                    id="FileGroup",
                ),
                id="File",
            ),
            id="MenuBar",
        )
        self.assertActionElementsEqual(actual, desired)

    def test_absolute_and_before_after(self):
        """ Does specifying absolute_position along with before, after work?
        """
        schema = MenuBarSchema(
            MenuSchema(
                GroupSchema(self.action1, self.action2, id="FileGroup"),
                id="File",
            )
        )
        extras = [
            SchemaAddition(
                factory=lambda: self.action3,
                id="action3",
                after="action2",
                path="MenuBar/File/FileGroup",
            ),
            SchemaAddition(
                factory=lambda: self.action4,
                after="action3",
                path="MenuBar/File/FileGroup",
            ),
            SchemaAddition(
                factory=lambda: self.action5,
                id="action5",
                absolute_position="last",
                path="MenuBar/File/FileGroup",
            ),
            SchemaAddition(
                factory=lambda: self.action6,
                absolute_position="last",
                before="action5",
                path="MenuBar/File/FileGroup",
            ),
        ]
        builder = TaskActionManagerBuilder(
            task=Task(menu_bar=schema, extra_actions=extras)
        )
        actual = builder.create_menu_bar_manager()
        desired = MenuBarManager(
            MenuManager(
                Group(
                    self.action1,
                    self.action2,
                    self.action3,
                    self.action4,
                    self.action6,
                    self.action5,
                    id="FileGroup",
                ),
                id="File",
            ),
            id="MenuBar",
        )
        self.assertActionElementsEqual(actual, desired)


if __name__ == "__main__":
    unittest.main()



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from itertools import count
from unittest import TestCase
from unittest.mock import Mock

from pyface.data_view.exporters.row_exporter import RowExporter
from pyface.data_view.i_data_wrapper import DataFormat


trivial_format = DataFormat(
    'null/null',
    lambda x: str(x).encode('utf-8'),
    lambda x: x,
)


class TestRowExporter(TestCase):

    def setUp(self):
        self.value_type = Mock()
        self.value_type.has_text = Mock(return_value=False)
        self.value_type.has_editor_value = Mock(return_value=True)
        self.value_type.get_editor_value = Mock(
            return_value=1,
            side_effect=count(),
        )

        self.model = Mock()
        self.model.get_column_count = Mock(return_value=3)
        self.model.get_value = Mock(return_value=0)
        self.model.get_value_type = Mock(return_value=self.value_type)

    def test_get_data(self):
        exporter = RowExporter(format=trivial_format)

        result = exporter.get_data(self.model, [((0,), (0,)), ((1,), ())])

        self.assertEqual(result, [[0, 1, 2], [3, 4, 5]])

    def test_get_data_deduplicate(self):
        exporter = RowExporter(format=trivial_format)

        result = exporter.get_data(
            self.model,
            [
                ((0,), (0,)),
                ((0,), (2,)),
                ((1,), ()),
            ])

        self.assertEqual(result, [[0, 1, 2], [3, 4, 5]])

    def test_get_data_row_headers(self):
        exporter = RowExporter(format=trivial_format, row_headers=True)

        result = exporter.get_data(self.model, [((0,), (0,)), ((1,), ())])

        self.assertEqual(result, [[0, 1, 2, 3], [4, 5, 6, 7]])

    def test_get_data_column_headers(self):
        exporter = RowExporter(format=trivial_format, column_headers=True)

        result = exporter.get_data(self.model, [((0,), (0,)), ((1,), ())])

        self.assertEqual(result, [[0, 1, 2], [3, 4, 5], [6, 7, 8]])



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


import importlib
import os
import sys

QtAPIs = [
    ("pyside2", "PySide2"),
    ("pyqt5", "PyQt5"),
    ("pyqt", "PyQt4"),
]


qt_api = None

# have we already imported a Qt API?
for api_name, module in QtAPIs:
    if module in sys.modules:
        qt_api = api_name
        break
else:
    # does our environment give us a preferred API?
    qt_api = os.environ.get("QT_API")

# if we have no preference, is a Qt API available? Or fail with ImportError.
if qt_api is None:
    for api_name, module in QtAPIs:
        try:
            importlib.import_module(module)
            importlib.import_module(".QtCore", module)
            qt_api = api_name
            break
        except ImportError:
            continue
    else:
        raise ImportError("Cannot import PySide2, PyQt5 or PyQt4")

# otherwise check QT_API value is valid
elif qt_api not in {api_name for api_name, module in QtAPIs}:
    msg = (
        "Invalid Qt API %r, valid values are: "
        + "'pyside2', 'pyqt' or 'pyqt5'"
    ) % qt_api
    raise RuntimeError(msg)

# useful constants
is_qt4 = qt_api in {"pyqt"}
is_qt5 = qt_api in {"pyqt5", "pyside2"}



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" The model for list boxes. """


from traits.api import Event, HasTraits


# Classes for event traits.
class ListModelEvent(object):
    """ Information about list model changes. """


class ListBoxModel(HasTraits):
    """ The model for list boxes. """

    # Events ----

    #: Fired when the contents of the list have changed.
    list_changed = Event()

    def get_item_count(self):
        """ Get the number of items in the list.

        Returns
        -------
        item_count : int
            The number of items in the list.
        """
        raise NotImplementedError()

    def get_item_at(self, index):
        """ Returns the item at the specified index.

        Parameters
        ----------
        index : int
            The index to return the value of.

        Returns
        -------
        label, item : str, any
            The user-visible string and model data of the item.
        """
        raise NotImplementedError()

    def fire_list_changed(self):
        """ Invoke this method when the list has changed. """
        self.list_changed = ListModelEvent()



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
from pyface.i_drop_handler import IDropHandler
from traits.api import Callable, HasTraits, List, provides, Str


@provides(IDropHandler)
class BaseDropHandler(HasTraits):
    """ Basic drop handler
    """

    # BaseDropHandler interface ---------------------------------------------

    #: Returns True if the current drop handler can handle the given drag event
    #: occurring on the given target widget.
    on_can_handle = Callable

    #: Performs drop action when drop event occurs on target widget.
    on_handle = Callable

    # IDropHandler interface ------------------------------------------------

    def can_handle_drop(self, event, target):
        return self.on_can_handle(event, target)

    def handle_drop(self, event, target):
        return self.on_handle(event, target)


@provides(IDropHandler)
class FileDropHandler(HasTraits):
    """ Class to handle backward compatible file drop events
    """

    # FileDropHandler interface ---------------------------------------------

    #: supported extensions
    extensions = List(Str)

    #: Called when file is opened. Takes single argument: path of file
    open_file = Callable

    # IDropHandler interface ------------------------------------------------

    def can_handle_drop(self, event, target):
        """ Does the drop event contails file data with matching extensions """
        if event.mimeData().hasUrls():
            for url in event.mimeData().urls():
                file_path = url.toLocalFile()
                if file_path.endswith(tuple(self.extensions)):
                    return True
        return False

    def handle_drop(self, event, target):
        """ Open the file using the supplied callback """
        for url in event.mimeData().urls():
            self.open_file(url.toLocalFile())



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


import unittest

from traits.api import Enum, HasTraits
from traits.testing.api import UnittestTools

from pyface.gui import GUI
from pyface.toolkit import toolkit
from pyface.util.testing import has_traitsui
from pyface.window import Window
from ..traitsui_widget_action import TraitsUIWidgetAction


@unittest.skipIf(not has_traitsui(), "TraitsUI not installed")
@unittest.skipIf(toolkit.toolkit == "wx", "wxPython not supported")
class TestTraitsUIWidgetAction(unittest.TestCase, UnittestTools):
    def setUp(self):
        self.gui = GUI()
        self.parent = Window()
        self.parent._create()
        self.parent.open()
        self.addCleanup(self._destroy_parent)
        self.gui.process_events()

    def _destroy_parent(self):
        self.parent.destroy()
        self.gui.process_events()
        self.parent = None

    def create_model(self):
        from traitsui.api import View, Item

        class SimpleEnum(HasTraits):
            value = Enum("a", "b", "c")
            view = View(Item("value"))

        return SimpleEnum()

    def test_traitsui_widget_action(self):
        from traitsui.api import View, Item

        class SimpleEnumAction(TraitsUIWidgetAction):
            value = Enum("a", "b", "c")
            view = View(Item("value"))

        action = SimpleEnumAction(name="Simple")
        control = action.create_control(self.parent.control)
        self.gui.process_events()

        editor = control._ui.get_editors("value")[0]

        with self.assertTraitChanges(action, "value", count=1):
            if toolkit.toolkit in {"qt", "qt4"}:
                editor.control.setCurrentIndex(1)
                editor.control.activated.emit(1)
            elif toolkit.toolkit == "wx":
                import wx

                event = wx.CommandEvent(
                    wx.EVT_CHOICE.typeId, editor.control.GetId()
                )
                event.SetString("b")
                wx.PostEvent(editor.control.GetEventHandler(), event)
            else:
                self.skipTest("Unknown toolkit")
            self.gui.process_events()

        self.assertEqual(action.value, "b")

    def test_traitsui_widget_action_model(self):
        from traitsui.api import View, Item

        class SimpleEnumAction(TraitsUIWidgetAction):
            view = View(Item("value"))

        model = self.create_model()
        action = SimpleEnumAction(name="Simple", model=model)
        control = action.create_control(self.parent.control)
        self.gui.process_events()

        editor = control._ui.get_editors("value")[0]

        with self.assertTraitChanges(model, "value", count=1):
            if toolkit.toolkit in {"qt", "qt4"}:
                editor.control.setCurrentIndex(1)
                editor.control.activated.emit(1)
            elif toolkit.toolkit == "wx":
                import wx

                event = wx.CommandEvent(
                    wx.EVT_CHOICE.typeId, editor.control.GetId()
                )
                event.SetString("b")
                wx.PostEvent(editor.control.GetEventHandler(), event)
            else:
                self.skipTest("Unknown toolkit")
            self.gui.process_events()

        self.assertEqual(model.value, "b")

    def test_traitsui_widget_action_model_view(self):
        from traitsui.api import HGroup, View, Item

        class ComplexEnumAction(TraitsUIWidgetAction):
            value = Enum("a", "b", "c")

            view = View(HGroup(Item("value"), Item("action.value")))

        model = self.create_model()
        action = ComplexEnumAction(name="Simple", model=model)
        control = action.create_control(self.parent.control)
        self.gui.process_events()

        editor = control._ui.get_editors("value")[0]

        with self.assertTraitChanges(model, "value", count=1):
            if toolkit.toolkit in {"qt", "qt4"}:
                editor.control.setCurrentIndex(1)
                editor.control.activated.emit(1)
            elif toolkit.toolkit == "wx":
                import wx

                event = wx.CommandEvent(
                    wx.EVT_CHOICE.typeId, editor.control.GetId()
                )
                event.SetString("b")
                wx.PostEvent(editor.control.GetEventHandler(), event)
            else:
                self.skipTest("Unknown toolkit")
            self.gui.process_events()

        self.assertEqual(model.value, "b")

        editor = control._ui.get_editors("value")[1]

        with self.assertTraitChanges(action, "value", count=1):
            if toolkit.toolkit in {"qt", "qt4"}:
                editor.control.setCurrentIndex(2)
                editor.control.activated.emit(2)
            elif toolkit.toolkit == "wx":
                event = wx.CommandEvent(
                    wx.EVT_CHOICE.typeId, editor.control.GetId()
                )
                event.SetString("c")
                wx.PostEvent(editor.control.GetEventHandler(), event)
            else:
                self.skipTest("Unknown toolkit")
            self.gui.process_events()

        self.assertEqual(action.value, "c")



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


import unittest

from traits.etsconfig.api import ETSConfig

from ..clipboard import clipboard

is_wx = (ETSConfig.toolkit == 'wx')


class TestObject(object):
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __eq__(self, other):
        if isinstance(other, TestObject):
            return all(
                getattr(other, key) == value
                for key, value in self.__dict__.items()
            )


class TestClipboard(unittest.TestCase):
    def setUp(self):
        self.clipboard = clipboard

    def test_set_text_data(self):
        self.clipboard.data = "test"
        self.assertTrue(self.clipboard.has_data)
        self.assertEqual(self.clipboard.data_type, "str")
        self.assertEqual(self.clipboard.data, "test")
        self.assertTrue(self.clipboard.has_text_data)
        self.assertEqual(self.clipboard.text_data, "test")
        self.assertFalse(self.clipboard.has_file_data)
        self.assertFalse(self.clipboard.has_object_data)

    def test_set_text_data_unicode(self):
        self.clipboard.data = "test"
        self.assertTrue(self.clipboard.has_data)
        self.assertEqual(self.clipboard.data_type, "str")
        self.assertEqual(self.clipboard.data, "test")
        self.assertTrue(self.clipboard.has_text_data)
        self.assertEqual(self.clipboard.text_data, "test")
        self.assertFalse(self.clipboard.has_file_data)
        self.assertFalse(self.clipboard.has_object_data)

    @unittest.skip("backends not consistent")
    def test_set_file_data(self):
        self.clipboard.data = ["file:///images"]
        self.assertTrue(self.clipboard.has_data)
        self.assertEqual(self.clipboard.data_type, "file")
        self.assertEqual(self.clipboard.data, ["/images"])
        self.assertTrue(self.clipboard.has_file_data)
        self.assertEqual(self.clipboard.file_data, ["/images"])
        self.assertFalse(self.clipboard.has_text_data)
        self.assertFalse(self.clipboard.has_object_data)

    def test_set_object_data(self):
        data = TestObject(foo="bar", baz=1)
        self.clipboard.data = data
        self.assertTrue(self.clipboard.has_data)
        self.assertEqual(self.clipboard.data_type, TestObject)
        self.assertEqual(self.clipboard.data, data)
        self.assertTrue(self.clipboard.has_object_data)
        self.assertEqual(self.clipboard.object_type, TestObject)
        self.assertEqual(self.clipboard.object_data, data)
        self.assertFalse(self.clipboard.has_text_data)
        self.assertFalse(self.clipboard.has_file_data)



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
import unittest

from traits.etsconfig.api import ETSConfig

# Skip tests if Enaml is not installed or we're using the wx backend.
SKIP_REASON = None
if ETSConfig.toolkit not in ["", "qt4"]:
    SKIP_REASON = "Enaml does not support WX"
else:
    try:
        from enaml.widgets.api import Label
        from traits_enaml.testing.gui_test_assistant import GuiTestAssistant
    except ImportError:
        SKIP_REASON = "traits_enaml is not installed"

if SKIP_REASON is not None:
    # Dummy class so that the TestEnamlTaskPane class definition below
    # doesn't fail.

    class GuiTestAssistant(object):
        pass


from pyface.tasks.api import EnamlDockPane, Task


class DummyDockPane(EnamlDockPane):
    def create_component(self):
        return Label(text="test label")


@unittest.skipIf(SKIP_REASON is not None, SKIP_REASON)
class TestEnamlDockPane(GuiTestAssistant, unittest.TestCase):

    # ------------------------------------------------------------------------
    # 'TestCase' interface
    # ------------------------------------------------------------------------

    def setUp(self):
        GuiTestAssistant.setUp(self)
        self.dock_pane = DummyDockPane(task=Task(id="dummy_task"))
        with self.event_loop():
            self.dock_pane.create(None)

    def tearDown(self):
        if self.dock_pane.control is not None:
            with self.delete_widget(self.dock_pane.control):
                self.dock_pane.destroy()
        del self.dock_pane
        GuiTestAssistant.tearDown(self)

    # ------------------------------------------------------------------------
    # Tests
    # ------------------------------------------------------------------------

    def test_creation(self):
        self.assertIsInstance(self.dock_pane.component, Label)
        self.assertIsNotNone(self.dock_pane.control)

    def test_destroy(self):
        dock_pane = self.dock_pane
        with self.delete_widget(dock_pane.control):
            dock_pane.destroy()
        self.assertIsNone(dock_pane.control)
        # Second destruction is a no-op.
        dock_pane.destroy()



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Helper functions to automatically generate unique IDs. """


from weakref import WeakKeyDictionary


class _ObjectCounter(object):
    """ Counts objects. """

    def __init__(self):
        self._objects_registry = WeakKeyDictionary()

    def get_count(self, obj):
        """ Return the number of times an object was seen.

        Objects must be hashable.

        """

        if obj in self._objects_registry:
            count = self._objects_registry[obj]
        else:
            count = 0

        return count

    def next_count(self, obj):
        """ Increase and return the number of times an object was seen.

        Objects must be hashable.

        """

        count = self.get_count(obj)
        self._objects_registry[obj] = count + 1
        return self._objects_registry[obj]


# Global object counter.
object_counter = _ObjectCounter()


def get_unique_id(object):
    """ Return a unique ID of the form ClassName_X, where X is an integer.

    It is only guaranteed that IDs are unique to a specific Python session, not
    across sessions.

    """

    class_ = object.__class__
    name = class_.__name__
    number = object_counter.next_count(class_)

    return name + "_" + str(number)



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
import unittest

from pyface.timer.api import CallbackTimer
from pyface.ui.qt4.util.gui_test_assistant import GuiTestAssistant
from traits.api import Event, HasStrictTraits


class ExampleObject(HasStrictTraits):
    """ Test class; target for test_event_loop_until_traits_change. """

    spam = Event()
    eggs = Event()


class TestGuiTestAssistant(GuiTestAssistant, unittest.TestCase):
    def test_event_loop_until_traits_change_with_single_trait_success(self):
        # event_loop_until_traits_change calls self.fail on timeout.
        obj = ExampleObject()

        # Successful case.
        with self.event_loop_until_traits_change(obj, "spam"):
            obj.spam = True

    def test_event_loop_until_traits_change_with_single_trait_failure(self):
        # event_loop_until_traits_change calls self.fail on timeout.
        obj = ExampleObject()

        # Failing case.
        with self.assertRaises(AssertionError):
            with self.event_loop_until_traits_change(obj, "spam", timeout=0.1):
                obj.eggs = True

    def test_event_loop_until_traits_change_with_multiple_traits_success(self):
        # event_loop_until_traits_change calls self.fail on timeout.
        obj = ExampleObject()
        with self.event_loop_until_traits_change(obj, "spam", "eggs"):
            obj.spam = True
            obj.eggs = True

    def test_event_loop_until_traits_change_with_multiple_traits_failure(self):
        # event_loop_until_traits_change calls self.fail on timeout.
        obj = ExampleObject()
        with self.assertRaises(AssertionError):
            with self.event_loop_until_traits_change(
                obj, "spam", "eggs", timeout=0.1
            ):
                obj.eggs = True

        # event_loop_until_traits_change calls self.fail on timeout.
        with self.assertRaises(AssertionError):
            with self.event_loop_until_traits_change(
                obj, "spam", "eggs", timeout=0.1
            ):
                obj.spam = True

    def test_event_loop_until_traits_change_with_no_traits_success(self):
        # event_loop_until_traits_change calls self.fail on timeout.
        obj = ExampleObject()

        # Successful case.
        with self.event_loop_until_traits_change(obj):
            pass

    def test_assert_eventually_true_in_gui_success(self):
        my_list = []

        timer = CallbackTimer(
            interval=0.05, callback=my_list.append, args=("bob",), repeat=1
        )

        timer.start()
        try:
            self.assertEventuallyTrueInGui(lambda: len(my_list) > 0)
            self.assertEqual(my_list, ["bob"])
        finally:
            timer.stop()

    def test_assert_eventually_true_in_gui_already_true(self):
        my_list = ["bob"]
        self.assertEventuallyTrueInGui(lambda: len(my_list) > 0)

    def test_assert_eventually_true_in_gui_failure(self):
        my_list = []
        with self.assertRaises(AssertionError):
            self.assertEventuallyTrueInGui(
                lambda: len(my_list) > 0, timeout=0.1
            )



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from types import MappingProxyType
import unittest

from traits.api import TraitError
from traits.testing.api import UnittestTools

from pyface.data_view.abstract_data_model import DataViewSetError
from pyface.data_view.value_types.api import IntValue, TextValue
from pyface.data_view.data_models.data_accessors import (
    AttributeDataAccessor,
    ConstantDataAccessor,
    IndexDataAccessor,
    KeyDataAccessor,
)


class AttributeDummy:

    def __init__(self, attr_value):
        self.attr_value = attr_value


class DataAccessorMixin(UnittestTools):

    def accessor_observer(self, event):
        self.accessor_event = event

    def test_title_type_changed(self):
        accessor = self.create_accessor()
        accessor.observe(self.accessor_observer, 'updated')

        with self.assertTraitChanges(accessor, 'updated', count=1):
            accessor.title_type = TextValue()

        self.assertEqual(self.accessor_event.new, (accessor, 'title'))

    def test_title_type_updated(self):
        accessor = self.create_accessor()
        accessor.observe(self.accessor_observer, 'updated')

        with self.assertTraitChanges(accessor, 'updated', count=1):
            accessor.title_type.updated = True

        self.assertEqual(self.accessor_event.new, (accessor, 'title'))

    def test_value_type_changed(self):
        accessor = self.create_accessor()
        accessor.observe(self.accessor_observer, 'updated')

        with self.assertTraitChanges(accessor, 'updated', count=1):
            accessor.value_type = TextValue()

        self.assertEqual(self.accessor_event.new, (accessor, 'value'))

    def test_value_type_updated(self):
        accessor = self.create_accessor()
        accessor.observe(self.accessor_observer, 'updated')

        with self.assertTraitChanges(accessor, 'updated', count=1):
            accessor.value_type.updated = True

        self.assertEqual(self.accessor_event.new, (accessor, 'value'))


class TestConstantDataAccessor(unittest.TestCase, DataAccessorMixin):

    def create_accessor(self):
        return ConstantDataAccessor(
            title='Test',
            value='test',
            value_type=TextValue(),
        )

    def test_defaults(self):
        accessor = ConstantDataAccessor()

        self.assertEqual(accessor.value, None)
        self.assertIsNone(accessor.value_type)
        self.assertIsInstance(accessor.title_type, TextValue)
        self.assertEqual(accessor.title, '')

    def test_typical_defaults(self):
        accessor = self.create_accessor()

        self.assertIsInstance(accessor.title_type, TextValue)
        self.assertEqual(accessor.title, 'Test')

    def test_get_value(self):
        accessor = self.create_accessor()
        obj = object()

        value = accessor.get_value(obj)

        self.assertEqual(value, 'test')

    def test_can_set_value(self):
        accessor = self.create_accessor()
        obj = object()

        can_set = accessor.can_set_value(obj)

        self.assertFalse(can_set)

    def test_set_value_error(self):
        accessor = self.create_accessor()
        obj = object()

        with self.assertRaises(DataViewSetError):
            accessor.set_value(obj, 'new_value')

    def test_value_updated(self):
        accessor = self.create_accessor()
        accessor.observe(self.accessor_observer, 'updated')

        with self.assertTraitChanges(accessor, 'updated', count=1):
            accessor.value = 'other_value'

        self.assertEqual(self.accessor_event.new, (accessor, 'value'))


class TestAttributeDataAccessor(unittest.TestCase, DataAccessorMixin):

    def create_accessor(self):
        return AttributeDataAccessor(
            attr='attr_value',
            value_type=TextValue(),
        )

    def test_defaults(self):
        accessor = AttributeDataAccessor()

        self.assertEqual(accessor.attr, '')
        self.assertIsNone(accessor.value_type)
        self.assertIsInstance(accessor.title_type, TextValue)
        self.assertEqual(accessor.title, '')

    def test_typical_defaults(self):
        accessor = self.create_accessor()

        self.assertIsInstance(accessor.title_type, TextValue)
        self.assertEqual(accessor.title, 'Attr Value')

    def test_get_value(self):
        accessor = self.create_accessor()
        obj = AttributeDummy('test_value')

        value = accessor.get_value(obj)

        self.assertEqual(value, 'test_value')

    def test_get_value_extended(self):
        accessor = self.create_accessor()
        accessor.attr = 'attr_value.attr_value'
        obj = AttributeDummy(AttributeDummy('test_value'))

        value = accessor.get_value(obj)

        self.assertEqual(value, 'test_value')

    def test_get_value_missing(self):
        accessor = self.create_accessor()
        accessor.attr = ''
        obj = AttributeDummy('test_value')

        with self.assertRaises(AttributeError):
            accessor.get_value(obj)

    def test_get_value_error(self):
        accessor = self.create_accessor()
        accessor.attr = 'other_attr'
        obj = AttributeDummy('test_value')

        with self.assertRaises(AttributeError):
            accessor.get_value(obj)

    def test_can_set_value(self):
        accessor = self.create_accessor()
        obj = AttributeDummy('test_value')

        can_set = accessor.can_set_value(obj)

        self.assertTrue(can_set)

    def test_can_set_value_false(self):
        accessor = AttributeDataAccessor()
        obj = AttributeDummy('test_value')

        can_set = accessor.can_set_value(obj)

        self.assertFalse(can_set)

    def test_set_value(self):
        accessor = self.create_accessor()
        obj = AttributeDummy('test_value')

        accessor.set_value(obj, 'new_value')

        self.assertEqual(obj.attr_value, 'new_value')

    def test_set_value_extended(self):
        accessor = self.create_accessor()
        accessor.attr = 'attr_value.attr_value'
        obj = AttributeDummy(AttributeDummy('test_value'))

        accessor.set_value(obj, 'new_value')

        self.assertEqual(obj.attr_value.attr_value, 'new_value')

    def test_set_value_error(self):
        accessor = AttributeDataAccessor()
        obj = AttributeDummy('test_value')

        with self.assertRaises(DataViewSetError):
            accessor.set_value(obj, 'new_value')

    def test_attr_updated(self):
        accessor = self.create_accessor()
        accessor.observe(self.accessor_observer, 'updated')

        with self.assertTraitChanges(accessor, 'updated', count=1):
            accessor.attr = 'other_attr'

        self.assertEqual(self.accessor_event.new, (accessor, 'value'))


class TestIndexDataAccessor(unittest.TestCase, DataAccessorMixin):

    def create_accessor(self):
        return IndexDataAccessor(
            index=1,
            value_type=TextValue(),
        )

    def test_defaults(self):
        accessor = IndexDataAccessor()

        self.assertEqual(accessor.index, 0)
        self.assertIsNone(accessor.value_type)
        self.assertIsInstance(accessor.title_type, TextValue)
        self.assertEqual(accessor.title, '0')

    def test_typical_defaults(self):
        accessor = self.create_accessor()

        self.assertIsInstance(accessor.title_type, TextValue)
        self.assertEqual(accessor.title, '1')

    def test_get_value(self):
        accessor = self.create_accessor()
        obj = ['zero', 'one', 'two', 'three']

        value = accessor.get_value(obj)

        self.assertEqual(value, 'one')

    def test_get_value_out_of_bounds(self):
        accessor = self.create_accessor()
        accessor.index = 10
        obj = ['zero', 'one', 'two', 'three']

        with self.assertRaises(IndexError):
            accessor.get_value(obj)

    def test_can_set_value(self):
        accessor = self.create_accessor()
        obj = ['zero', 'one', 'two', 'three']

        can_set = accessor.can_set_value(obj)

        self.assertTrue(can_set)

    def test_can_set_value_false(self):
        accessor = self.create_accessor()
        obj = ['zero']

        can_set = accessor.can_set_value(obj)

        self.assertFalse(can_set)

    def test_can_set_value_immuatble(self):
        accessor = self.create_accessor()
        obj = ('zero', 'one', 'two', 'three')

        can_set = accessor.can_set_value(obj)

        self.assertFalse(can_set)

    def test_set_value(self):
        accessor = self.create_accessor()
        obj = ['zero', 'one', 'two', 'three']

        accessor.set_value(obj, 'new_value')

        self.assertEqual(obj[1], 'new_value')

    def test_set_value_error(self):
        accessor = self.create_accessor()
        obj = ['zero']

        with self.assertRaises(DataViewSetError):
            accessor.set_value(obj, 'new_value')

    def test_index_updated(self):
        accessor = self.create_accessor()
        accessor.observe(self.accessor_observer, 'updated')

        with self.assertTraitChanges(accessor, 'updated', count=1):
            accessor.index = 2

        self.assertEqual(self.accessor_event.new, (accessor, 'value'))


class TestKeyDataAccessor(unittest.TestCase, DataAccessorMixin):

    def create_accessor(self):
        return KeyDataAccessor(
            key='one',
            value_type=TextValue(),
        )

    def test_defaults(self):
        accessor = KeyDataAccessor()

        self.assertIsNone(accessor.key)
        self.assertIsNone(accessor.value_type)
        self.assertIsInstance(accessor.title_type, TextValue)
        self.assertEqual(accessor.title, 'None')

    def test_typical_defaults(self):
        accessor = self.create_accessor()

        self.assertIsInstance(accessor.title_type, TextValue)
        self.assertEqual(accessor.title, 'One')

    def test_unhashable_error(self):
        accessor = self.create_accessor()
        with self.assertRaises(TraitError):
            accessor.key = []

    def test_get_value(self):
        accessor = self.create_accessor()
        obj = {'one': 'a', 'two': 'b'}

        value = accessor.get_value(obj)

        self.assertEqual(value, 'a')

    def test_get_value_missing(self):
        accessor = self.create_accessor()
        accessor.key = 'three'
        obj = {'one': 'a', 'two': 'b'}

        with self.assertRaises(KeyError):
            accessor.get_value(obj)

    def test_can_set_value(self):
        accessor = self.create_accessor()
        obj = {'one': 'a', 'two': 'b'}

        can_set = accessor.can_set_value(obj)

        self.assertTrue(can_set)

    def test_can_set_value_new(self):
        accessor = self.create_accessor()
        accessor.key = 'three'
        obj = {'one': 'a', 'two': 'b'}

        can_set = accessor.can_set_value(obj)

        self.assertTrue(can_set)

    def test_can_set_value_immutable(self):
        accessor = self.create_accessor()
        # TODO: eventually replace with frozenmap in 3.9
        obj = MappingProxyType({'one': 'a', 'two': 'b'})

        can_set = accessor.can_set_value(obj)

        self.assertFalse(can_set)

    def test_set_value(self):
        accessor = self.create_accessor()
        obj = {'one': 'a', 'two': 'b'}

        accessor.set_value(obj, 'new_value')

        self.assertEqual(obj['one'], 'new_value')

    def test_set_value_new(self):
        accessor = self.create_accessor()
        accessor.key = 'three'
        obj = {'one': 'a', 'two': 'b'}

        accessor.set_value(obj, 'new_value')

        self.assertEqual(obj['three'], 'new_value')

    def test_set_value_error(self):
        accessor = KeyDataAccessor()
        accessor.key = 'one'
        obj = MappingProxyType({'one': 'a', 'two': 'b'})

        with self.assertRaises(DataViewSetError):
            accessor.set_value(obj, 'new_value')

    def test_key_updated(self):
        accessor = self.create_accessor()
        accessor.observe(self.accessor_observer, 'updated')

        with self.assertTraitChanges(accessor, 'updated', count=1):
            accessor.key = 2

        self.assertEqual(self.accessor_event.new, (accessor, 'value'))



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from traits.api import Callable, HasTraits, Str, Enum


class SchemaAddition(HasTraits):
    """ An addition to an existing menu bar or tool bar schema.
    """

    #: The schema addition's identifier. This optional, but if left
    #: unspecified, other schema additions will be unable to refer to this one.
    id = Str()

    #: A callable to create the item. Should have the following signature:
    #:    callable() -> Action | ActionItem | Group | MenuManager |
    #:                  GroupSchema | MenuSchema
    #:
    #: If the result is a schema, it will itself admit of extension by other
    #: additions. If not, the result will be fixed.
    factory = Callable

    #: A forward-slash-separated path through the action hierarchy to the menu
    #: to add the action, group or menu to. For example:
    #: - To add an item to the menu bar: ``path = "MenuBar"``
    #: - To add an item to the tool bar: ``path = "ToolBar"``
    #: - To add an item to a sub-menu: ``path = "MenuBar/File/New"``
    path = Str()

    #: The item appears after the item with this ID.
    #: - for groups, this is the ID of another group.
    #: - for menus and actions, this is the ID of another menu or action.
    after = Str()

    #: The action appears before the item with this ID.
    #: - for groups, this is the ID of another group.
    #: - for menus and actions, this is the ID of another menu or action.
    before = Str()

    #: The action appears at the absolute specified position first or last.
    #: This is useful for example to keep the File menu the first menu in a
    #: menubar, the help menu the last etc.
    #: If multiple actions in a schema have absolute_position 'first', they
    #: will appear in the same order specified; unless 'before' and 'after'
    #: traits are set to sort these multiple items.
    #: This trait takes precedence over 'after' and 'before', and values of
    #: those traits that are not compatible with the  absolute_position are
    #: ignored.
    absolute_position = Enum(None, "first", "last")



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from unittest import TestCase
from unittest.mock import Mock

from traits.testing.api import UnittestTools

from pyface.data_view.data_view_errors import DataViewSetError
from pyface.data_view.value_types.editable_value import EditableValue

class EditableWithValid(EditableValue):

    def is_valid(self, model, row, column, value):
        return value >= 0


class TestEditableValue(UnittestTools, TestCase):

    def setUp(self):
        self.model = Mock()
        self.model.get_value = Mock(return_value=1.0)
        self.model.can_set_value = Mock(return_value=True)
        self.model.set_value = Mock()

    def test_default(self):
        value_type = EditableValue()
        self.assertTrue(value_type.is_editable)

    def test_is_valid(self):
        value_type = EditableValue()
        result = value_type.is_valid(self.model, [0], [0], 2.0)
        self.assertTrue(result)

    def test_has_editor_value(self):
        value_type = EditableValue()
        result = value_type.has_editor_value(self.model, [0], [0])
        self.assertTrue(result)

    def test_has_editor_value_not_editable(self):
        value_type = EditableValue(is_editable=False)
        result = value_type.has_editor_value(self.model, [0], [0])
        self.assertFalse(result)

    def test_set_editor_value(self):
        value_type = EditableValue()
        value_type.set_editor_value(self.model, [0], [0], 2.0)
        self.model.set_value.assert_called_once_with([0], [0], 2.0)

    def test_set_editor_value_set_value_raises(self):
        self.model.set_value = Mock(side_effect=DataViewSetError)
        value_type = EditableValue(is_editable=False)
        with self.assertRaises(DataViewSetError):
            value_type.set_editor_value(self.model, [0], [0], 2.0)
        self.model.set_value.assert_called_once_with([0], [0], 2.0)

    def test_set_editor_value_not_valid(self):
        value_type = EditableWithValid()
        with self.assertRaises(DataViewSetError):
            value_type.set_editor_value(self.model, [0], [0], -1.0)
        self.model.set_value.assert_not_called()

    def test_is_editable_update(self):
        value_type = EditableValue()
        with self.assertTraitChanges(value_type, 'updated', count=1):
            value_type.is_editable = False



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" The preference dialog. """


import wx


from traits.api import Any, Dict, Float, Instance, Str


from pyface.preference.preference_node import PreferenceNode
from pyface.ui.wx.heading_text import HeadingText
from pyface.ui.wx.layered_panel import LayeredPanel
from pyface.ui.wx.split_dialog import SplitDialog
from pyface.ui.wx.viewer.tree_viewer import TreeViewer
from pyface.viewer.default_tree_content_provider import (
    DefaultTreeContentProvider,
)
from pyface.wx.util.font_helper import new_font_like


class PreferenceDialog(SplitDialog):
    """ The preference dialog. """

    # 'Dialog' interface ---------------------------------------------------

    # The dialog title.
    title = Str("Preferences")

    # 'SplitDialog' interface ---------------------------------------------#

    # The ratio of the size of the left/top pane to the right/bottom pane.
    ratio = Float(0.25)

    # 'PreferenceDialog' interface -----------------------------------------

    # The root of the preference hierarchy.
    root = Instance(PreferenceNode)

    # Private interface ----------------------------------------------------

    # The preference pages in the dialog (they are created lazily).
    _pages = Dict()

    # The current visible preference page.
    _current_page = Any()

    # ------------------------------------------------------------------------
    # Protected 'Dialog' interface.
    # ------------------------------------------------------------------------

    def _create_buttons(self, parent):
        """ Creates the buttons. """

        sizer = wx.BoxSizer(wx.HORIZONTAL)

        # 'Done' button.
        done = wx.Button(parent, wx.ID_OK, "Done")
        done.SetDefault()
        parent.Bind(wx.EVT_BUTTON, self._wx_on_ok, wx.ID_OK)
        sizer.Add(done)

        return sizer

    # ------------------------------------------------------------------------
    # Protected 'SplitDialog' interface.
    # ------------------------------------------------------------------------

    def _create_lhs(self, parent):
        """ Creates the panel containing the preference page tree. """

        return self._create_tree(parent)

    def _create_rhs(self, parent):
        """ Creates the panel containing the selected preference page. """

        panel = wx.Panel(parent, -1, style=wx.CLIP_CHILDREN)
        sizer = wx.BoxSizer(wx.VERTICAL)
        panel.SetSizer(sizer)
        panel.SetAutoLayout(True)

        # The 'pretty' title bar ;^)
        self.__title = HeadingText(panel)
        sizer.Add(
            self.__title.control,
            0,
            wx.EXPAND | wx.BOTTOM | wx.LEFT | wx.RIGHT,
            5,
        )

        # The preference page of the node currently selected in the tree.
        self._layered_panel = LayeredPanel(panel, min_width=-1, min_height=-1)
        sizer.Add(
            self._layered_panel.control,
            1,
            wx.EXPAND | wx.TOP | wx.LEFT | wx.RIGHT,
            5,
        )

        # The 'Restore Defaults' button etc.
        buttons = self._create_page_buttons(panel)
        sizer.Add(buttons, 0, wx.ALIGN_RIGHT | wx.TOP | wx.RIGHT, 5)

        # A separator.
        line = wx.StaticLine(panel, -1)
        sizer.Add(line, 0, wx.EXPAND | wx.TOP | wx.LEFT | wx.RIGHT, 5)

        # Resize the panel to fit the sizer's minimum size.
        sizer.Fit(panel)

        return panel

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _create_tree(self, parent):
        """ Creates the preference page tree. """

        tree_viewer = TreeViewer(
            parent,
            input=self.root,
            show_images=False,
            show_root=False,
            content_provider=DefaultTreeContentProvider(),
        )

        tree_viewer.observe(self._on_selection_changed, "selection")

        return tree_viewer.control

    def _create_page_buttons(self, parent):
        """ Creates the 'Restore Defaults' button, etc.

        At the moment the "etc." is an optional 'Help' button.

        """

        self._button_sizer = sizer = wx.BoxSizer(wx.HORIZONTAL)

        # 'Help' button. Comes first so 'Restore Defaults' doesn't jump around.
        self._help = help = wx.Button(parent, -1, "Help")
        parent.Bind(wx.EVT_BUTTON, self._on_help, help.GetId())
        sizer.Add(help, 0, wx.RIGHT, 5)

        # 'Restore Defaults' button.
        restore = wx.Button(parent, -1, "Restore Defaults")
        parent.Bind(wx.EVT_BUTTON, self._on_restore_defaults, restore.GetId())
        sizer.Add(restore)

        return sizer

    # ------------------------------------------------------------------------
    # wx event handlers.
    # ------------------------------------------------------------------------

    def _on_restore_defaults(self, event):
        """ Called when the 'Restore Defaults' button is pressed. """

        page = self._pages[self._layered_panel.current_layer_name]
        page.restore_defaults()

    def _on_help(self, event):
        """ Called when the 'Help' button is pressed. """

        page = self._pages[self._layered_panel.current_layer_name]
        page.show_help_topic()

        return

    # ------------------------------------------------------------------------
    # Trait event handlers.
    # ------------------------------------------------------------------------

    def _on_selection_changed(self, event):
        """ Called when a node in the tree is selected. """
        selection = event.new
        if len(selection) > 0:
            # The tree is in single selection mode.
            node = selection[0]

            # We only show the help button if the selected node has a help
            # topic Id.
            if len(node.help_id) > 0:
                self._button_sizer.Show(self._help, True)

            else:
                self._button_sizer.Show(self._help, False)

            # Show the selected preference page.
            layered_panel = self._layered_panel
            parent = self._layered_panel.control

            # If we haven't yet displayed the node's preference page during the
            # lifetime of this dialog, then we have to create it.
            if not layered_panel.has_layer(node.name):
                page = node.create_page()
                layered_panel.add_layer(node.name, page.create_control(parent))
                self._pages[node.name] = page

            layered_panel.show_layer(node.name)
            self.__title.text = node.name

        return



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import sys


from traits.api import provides


from pyface.tasks.i_editor_area_pane import IEditorAreaPane
from .editor_area_pane import EditorAreaPane

# ----------------------------------------------------------------------------
# 'AdvancedEditorAreaPane' class.
# ----------------------------------------------------------------------------


@provides(IEditorAreaPane)
class SplitEditorAreaPane(EditorAreaPane):
    """ The toolkit-specific implementation of an AdvancedEditorAreaPane.

    See the IAdvancedEditorAreaPane interface for API documentation.
    """

    # No additional functionality over the standard EditorAreaPane in wx yet.



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Dialog utilities. """


import wx

# A file dialog wildcard for Python files.
WILDCARD_PY = "Python files (*.py)|*.py|"

# A file dialog wildcard for text files.
WILDCARD_TXT = "Text files (*.txt)|*.txt|"

# A file dialog wildcard for all files.
WILDCARD_ALL = "All files (*.*)|*.*"

# A file dialog wildcard for Zip archives.
WILDCARD_ZIP = "Zip files (*.zip)|*.zip|"


class OpenFileDialog(wx.FileDialog):
    """ An open-file dialog. """

    def __init__(self, parent=None, **kw):
        """ Constructor. """

        style = wx.OPEN | wx.HIDE_READONLY

        # Base-class constructor.
        wx.FileDialog.__init__(self, parent, "Open", style=style, **kw)


class OpenDirDialog(wx.DirDialog):
    """ An open-directory dialog. """

    def __init__(self, parent=None, **kw):
        """ Constructor. """

        style = wx.OPEN | wx.HIDE_READONLY | wx.DD_NEW_DIR_BUTTON

        # Base-class constructor.
        wx.DirDialog.__init__(self, parent, "Open", style=style, **kw)


class SaveFileAsDialog(wx.FileDialog):
    """ A save-file dialog. """

    def __init__(self, parent=None, **kw):
        """ Constructor. """

        style = wx.SAVE | wx.OVERWRITE_PROMPT

        # Base-class constructor.
        wx.FileDialog.__init__(self, parent, "Save As", style=style, **kw)


def confirmation(parent, message, title=None, default=wx.NO_DEFAULT):
    """ Displays a confirmation dialog. """

    dialog = wx.MessageDialog(
        parent,
        message,
        _get_title(title, parent, "Confirmation"),
        wx.YES_NO | default | wx.ICON_EXCLAMATION | wx.STAY_ON_TOP,
    )

    result = dialog.ShowModal()
    dialog.Destroy()

    return result


def yes_no_cancel(parent, message, title=None, default=wx.NO_DEFAULT):
    """ Displays a Yes/No/Cancel dialog. """

    dialog = wx.MessageDialog(
        parent,
        message,
        _get_title(title, parent, "Confirmation"),
        wx.YES_NO | wx.CANCEL | default | wx.ICON_EXCLAMATION | wx.STAY_ON_TOP,
    )

    result = dialog.ShowModal()
    dialog.Destroy()

    return result


def information(parent, message, title=None):
    """ Displays a modal information dialog. """

    dialog = wx.MessageDialog(
        parent,
        message,
        _get_title(title, parent, "Information"),
        wx.OK | wx.ICON_INFORMATION | wx.STAY_ON_TOP,
    )

    dialog.ShowModal()
    dialog.Destroy()


def warning(parent, message, title=None):
    """ Displays a modal warning dialog. """

    dialog = wx.MessageDialog(
        parent,
        message,
        _get_title(title, parent, "Warning"),
        wx.OK | wx.ICON_EXCLAMATION | wx.STAY_ON_TOP,
    )

    dialog.ShowModal()
    dialog.Destroy()


def error(parent, message, title=None):
    """ Displays a modal error dialog. """

    dialog = wx.MessageDialog(
        parent,
        message,
        _get_title(title, parent, "Error"),
        wx.OK | wx.ICON_ERROR | wx.STAY_ON_TOP,
    )

    dialog.ShowModal()
    dialog.Destroy()


def _get_title(title, parent, default):
    """ Get a sensible title for a dialog! """

    if title is None:
        if parent is not None:
            title = parent.GetTitle()

        else:
            title = default

    return title



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

from traits.testing.api import UnittestTools

from pyface.undo.api import CommandStack, UndoManager
from pyface.undo.tests.testing_commands import SimpleCommand, UnnamedCommand

from pyface.undo.action.api import RedoAction, UndoAction


class TestRedoAction(unittest.TestCase):

    def setUp(self):
        self.stack = CommandStack()
        self.undo_manager = UndoManager()
        self.stack.undo_manager = self.undo_manager
        self.undo_manager.active_stack = self.stack

        self.command = SimpleCommand()

    def test_update(self):
        redo_action = RedoAction(command=self.command, undo_manager=self.undo_manager)
        self.stack.push(self.command)
        self.undo_manager.undo()
        self.assertTrue(redo_action.enabled)
        self.assertEqual(redo_action.name, "&Redo Increment by 1")


class TestUndoAction(unittest.TestCase):

    def setUp(self):
        self.stack = CommandStack()
        self.undo_manager = UndoManager()
        self.stack.undo_manager = self.undo_manager
        self.undo_manager.active_stack = self.stack

        self.command = SimpleCommand()

    def test_update(self):
        undo_action = UndoAction(command=self.command, undo_manager=self.undo_manager)
        self.stack.push(self.command)
        self.assertTrue(undo_action.enabled)
        self.assertEqual(undo_action.name, "&Undo Increment by 1")



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import csv
import json
from unittest import TestCase
from unittest.mock import MagicMock, patch

from traits.api import HasTraits, Int, Str
from traits.testing.optional_dependencies import numpy as np, requires_numpy
from traits.version import version

from pyface.data_view.data_formats import (
    from_csv, from_csv_column, from_csv_row, from_npy, from_json,
    to_csv, to_csv_column, to_csv_row, to_npy, to_json,
)


def default_extractor(obj):
    return obj.__getstate__()


def object_hook(data):
    obj = ExampleObject()
    obj.__setstate__(data)
    return obj


class ExampleObject(HasTraits):
    a = Int(1)
    b = Str("two")
    c = Str("three", transient=True)


class TestJSON(TestCase):

    def test_to_json(self):
        test_data = {'a': 1, 'b': 'two'}

        raw_data = to_json(test_data)

        self.assertEqual(raw_data, b'{"a":1,"b":"two"}')

    def test_to_json_default(self):
        test_data = ExampleObject()

        raw_data = to_json(test_data, default=default_extractor)

        self.assertEqual(
            raw_data,
            b'{"a":1,"b":"two","__traits_version__":"'
            + version.encode('utf-8')
            + b'"}'
        )

    def test_from_json(self):
        raw_data = b'{"a":1,"b":"two"}'

        data = from_json(raw_data)

        self.assertEqual(data, {'a': 1, 'b': 'two'})

    def test_from_json_object_hook(self):
        raw_data = (
            b'{"a":2,"b":"four","__traits_version__":"'
            + version.encode('utf-8')
            + b'"}'
        )

        data = from_json(raw_data, object_hook=object_hook)

        self.assertIsInstance(data, ExampleObject)
        self.assertEqual(data.a, 2)
        self.assertEqual(data.b, 'four')
        self.assertEqual(data.c, 'three')


class TestCSV(TestCase):

    def test_to_csv(self):
        test_data = [['one', 2], ['three,four', 5]]

        raw_data = to_csv(test_data)

        self.assertEqual(raw_data, b'one,2\r\n"three,four",5\r\n')

    def test_to_csv_delimiter(self):
        test_data = [['one', 2], ['three,four', 5]]

        raw_data = to_csv(test_data, delimiter='\t')

        self.assertEqual(raw_data, b'one\t2\r\nthree,four\t5\r\n')

    def test_to_csv_encoding(self):
        test_data = [['øne', 2], ['three,four', 5]]

        raw_data = to_csv(test_data, encoding='latin-1')

        self.assertEqual(raw_data, b'\xf8ne,2\r\n"three,four",5\r\n')

    def test_from_csv(self):
        raw_data = b'one,2\r\n"three,four",5\r\n'

        data = from_csv(raw_data)

        self.assertEqual(data, [['one', '2'], ['three,four', '5']])

    def test_from_csv_delimiter(self):
        raw_data = b'one\t2\r\nthree,four\t5\r\n'

        data = from_csv(raw_data, delimiter='\t')

        self.assertEqual(data, [['one', '2'], ['three,four', '5']])

    def test_from_csv_encoding(self):
        raw_data = b'\xf8ne,2\r\n"three,four",5\r\n'

        data = from_csv(raw_data, encoding='latin-1')

        self.assertEqual(data, [['øne', '2'], ['three,four', '5']])

    def test_to_csv_column(self):
        test_data = ['one', 2, 'three,four']

        raw_data = to_csv_column(test_data)

        self.assertEqual(raw_data, b'one\r\n2\r\n"three,four"\r\n')

    def test_to_csv_column_delimiter(self):
        test_data = ['one', 2, 'three,four']

        raw_data = to_csv_column(test_data, delimiter='\t')

        self.assertEqual(raw_data, b'one\r\n2\r\nthree,four\r\n')

    def test_to_csv_column_encoding(self):
        test_data = ['øne', 2, 'three,four']

        raw_data = to_csv_column(test_data, encoding='latin-1')

        self.assertEqual(raw_data, b'\xf8ne\r\n2\r\n"three,four"\r\n')

    def test_from_csv_column(self):
        raw_data = b'one\r\n2\r\n"three,four"\r\n'

        data = from_csv_column(raw_data)

        self.assertEqual(data, ['one', '2', 'three,four'])

    def test_from_csv_column_delimiter(self):
        raw_data = b'one\r\n2\r\nthree,four\r\n'

        data = from_csv_column(raw_data, delimiter='\t')

        self.assertEqual(data, ['one', '2', 'three,four'])

    def test_from_csv_column_encoding(self):
        raw_data = b'\xf8ne\r\n2\r\n"three,four"\r\n'

        data = from_csv_column(raw_data, encoding='latin-1')

        self.assertEqual(data, ['øne', '2', 'three,four'])

    def test_to_csv_row(self):
        test_data = ['one', 2, 'three,four']

        raw_data = to_csv_row(test_data)

        self.assertEqual(raw_data, b'one,2,"three,four"\r\n')

    def test_to_csv_row_delimiter(self):
        test_data = ['one', 2, 'three,four']

        raw_data = to_csv_row(test_data, delimiter='\t')

        self.assertEqual(raw_data, b'one\t2\tthree,four\r\n')

    def test_to_csv_row_encoding(self):
        test_data = ['øne', 2, 'three,four']

        raw_data = to_csv_row(test_data, encoding='latin-1')

        self.assertEqual(raw_data, b'\xf8ne,2,"three,four"\r\n')

    def test_from_csv_row(self):
        raw_data = b'one,2,"three,four"\r\n'

        data = from_csv_row(raw_data)

        self.assertEqual(data, ['one', '2', 'three,four'])

    def test_from_csv_row_delimiter(self):
        raw_data = b'one\t2\tthree,four\r\n'

        data = from_csv_row(raw_data, delimiter='\t')

        self.assertEqual(data, ['one', '2', 'three,four'])

    def test_from_csv_row_encoding(self):
        raw_data = b'\xf8ne,2,"three,four"\r\n'

        data = from_csv_row(raw_data, encoding='latin-1')

        self.assertEqual(data, ['øne', '2', 'three,four'])


@requires_numpy
class TestNpy(TestCase):

    def test_to_npy(self):
        data = np.array([[1, 2, 3], [4, 5, 6]], dtype='uint8')

        raw_data = to_npy(data)

        self.assertEqual(
            raw_data,
            b"\x93NUMPY\x01\x00v\x00{'descr': '|u1', 'fortran_order': False, 'shape': (2, 3), }                                                          \n"  # noqa: E501
            + b"\x01\x02\x03\x04\x05\x06"
        )

    def test_from_npy(self):
        raw_data = (
            b"\x93NUMPY\x01\x00v\x00{'descr': '|u1', 'fortran_order': False, 'shape': (2, 3), }                                                          \n"  # noqa: E501
            + b"\x01\x02\x03\x04\x05\x06"
        )

        data = from_npy(raw_data)

        np.testing.assert_array_equal(
            data,
            np.array([[1, 2, 3], [4, 5, 6]], dtype='uint8')
        )



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" The interface to system metrics (screen width and height etc). """


from traits.api import HasTraits, Int, Interface, Tuple


class ISystemMetrics(Interface):
    """ The interface to system metrics (screen width and height etc). """

    # 'ISystemMetrics' interface -------------------------------------------

    #: The width of the screen in pixels.
    screen_width = Int()

    #: The height of the screen in pixels.
    screen_height = Int()

    #: Background color of a standard dialog window as a tuple of RGB values
    #: between 0.0 and 1.0.
    # FIXME v3: Why isn't this a traits colour?
    dialog_background_color = Tuple()


class MSystemMetrics(HasTraits):
    """ The mixin class that contains common code for toolkit specific
    implementations of the ISystemMetrics interface.
    """



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
# (C) Copyright 2007 Riverbank Computing Limited
# This software is provided without warranty under the terms of the BSD license.
# However, when used with the GPL version of PyQt the additional terms described in the PyQt GPL exception also apply


""" The PyQt specific implementation of a menu bar manager. """


import sys


from pyface.qt import QtGui


from pyface.action.action_manager import ActionManager


class MenuBarManager(ActionManager):
    """ A menu bar manager realizes itself in errr, a menu bar control. """

    # ------------------------------------------------------------------------
    # 'MenuBarManager' interface.
    # ------------------------------------------------------------------------

    def create_menu_bar(self, parent, controller=None):
        """ Creates a menu bar representation of the manager. """

        # If a controller is required it can either be set as a trait on the
        # menu bar manager (the trait is part of the 'ActionManager' API), or
        # passed in here (if one is passed in here it takes precedence over the
        # trait).
        if controller is None:
            controller = self.controller

        # Create the menu bar. Work around disappearing menu bars on OS X
        # (particulary on PySide but also under certain circumstances on PyQt4).
        if isinstance(parent, QtGui.QMainWindow) and sys.platform == "darwin":
            parent.menuBar().setParent(None)
            menu_bar = parent.menuBar()
        else:
            menu_bar = QtGui.QMenuBar(parent)

        # Every item in every group must be a menu manager.
        for group in self.groups:
            for item in group.items:
                menu = item.create_menu(parent, controller)
                menu.menuAction().setText(item.name)
                menu_bar.addMenu(menu)

        return menu_bar



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" The Wx-specific implementation of the text field class """


from traits.api import Any, Instance, Str, provides

import wx

from pyface.fields.i_field import IField, MField
from pyface.ui.wx.widget import Widget


@provides(IField)
class Field(MField, Widget):
    """ The Wxspecific implementation of the field class

    This is an abstract class which is not meant to be instantiated.
    """

    #: The value held by the field.
    value = Any()

    #: A tooltip for the field.
    tooltip = Str()

    #: An optional context menu for the field.
    context_menu = Instance("pyface.action.menu_manager.MenuManager")

    # ------------------------------------------------------------------------
    # IField interface
    # ------------------------------------------------------------------------

    def _initialize_control(self):
        """ Perform any toolkit-specific initialization for the control. """
        self.control.SetToolTip(self.tooltip)
        self.control.Enable(self.enabled)
        self.control.Show(self.visible)

    # ------------------------------------------------------------------------
    # IWidget interface
    # ------------------------------------------------------------------------

    def _create(self):
        super(Field, self)._create()
        self._add_event_listeners()

    def destroy(self):
        self._remove_event_listeners()
        super(Field, self).destroy()

    # ------------------------------------------------------------------------
    # Private interface
    # ------------------------------------------------------------------------

    def _get_control_tooltip(self):
        """ Toolkit specific method to get the control's tooltip. """
        return self.control.GetToolTipText()

    def _set_control_tooltip(self, tooltip):
        """ Toolkit specific method to set the control's tooltip. """
        self.control.SetToolTip(tooltip)

    def _observe_control_context_menu(self, remove=False):
        """ Toolkit specific method to change the control menu observer. """
        if remove:
            self.control.Unbind(
                wx.EVT_CONTEXT_MENU, handler=self._handle_context_menu
            )
        else:
            self.control.Bind(wx.EVT_CONTEXT_MENU, self._handle_context_menu)

    def _handle_control_context_menu(self, event):
        """ Signal handler for displaying context menu. """
        if self.control is not None and self.context_menu is not None:
            menu = self.context_menu.create_menu(self.control)
            self.control.PopupMenu(menu)



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Heading text. """


from traits.api import HasTraits, Instance, Int, Interface, Str


from pyface.i_image_resource import IImageResource


class IHeadingText(Interface):
    """ Heading text. """

    # 'IHeadingText' interface ---------------------------------------------

    #: Heading level.
    #
    # fixme: Currently we ignore anything but one, but in future we could
    # have different visualizations based on the level.
    level = Int(1)

    #: The heading text.
    text = Str("Default")

    #: The background image.
    image = Instance(IImageResource)


class MHeadingText(HasTraits):
    """ The mixin class that contains common code for toolkit specific
    implementations of the IHeadingText interface.
    """



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Base class for all sorters. """

from functools import cmp_to_key

from traits.api import HasTraits


class Sorter(HasTraits):
    """ Abstract base class for all sorters. """

    # ------------------------------------------------------------------------
    # 'ViewerSorter' interface.
    # ------------------------------------------------------------------------

    def sort(self, widget, parent, nodes):
        """ Sorts a list of nodes IN PLACE.

        'widget' is the widget that we are sorting nodes for.
        'parent' is the parent node.
        'nodes'  is the list of nodes to sort.

        Returns the list that was sorted IN PLACE (for convenience).

        """

        # This creates a comparison function with the names 'widget' and
        # 'parent' bound to the corresponding arguments to this method.
        def key(node):
            """ Comparator. """

            return self.key(widget, parent, node)

        nodes.sort(key=key)

        return nodes

    def key(self, widget, parent, node):
        category = self.category(widget, parent, node)
        text = widget.model.get_text(node)
        return (category, text)

    def category(self, widget, parent, node):
        """ Returns the category (an integer) for an node.

        'parent' is the parent node.
        'nodes'  is the node to return the category for.

        Categories are used to sort nodes into bins.  The bins are arranged in
        ascending numeric order.  The nodes within a bin are arranged as
        dictated by the sorter's 'compare' method.

        By default all nodes are given the same category (0).

        """

        return 0

    def is_sorter_trait(self, node, trait_name):
        """ Is the sorter affected by changes to a node's trait?

        'node'       is the node.
        'trait_name' is the name of the trait.

        Returns True if the sorter would be affected by changes to the trait
        named 'trait_name' on the specified node.

        By default we return False.

        """

        return False



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
from . import qt_api

if qt_api == "pyqt":
    from PyQt4.QtNetwork import *

elif qt_api == "pyqt5":
    from PyQt5.QtNetwork import *

else:
    from PySide2.QtNetwork import *



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!


""" A tool bar manager realizes itself in a tool palette control.
"""


import wx


from traits.api import Any, Bool, Enum, Instance, Tuple


from pyface.image_cache import ImageCache
from pyface.action.action_manager import ActionManager
from .tool_palette import ToolPalette


class ToolPaletteManager(ActionManager):
    """ A tool bar manager realizes itself in a tool palette bar control. """

    # 'ToolPaletteManager' interface ---------------------------------------

    # The size of tool images (width, height).
    image_size = Tuple((16, 16))

    # Should we display the name of each tool bar tool under its image?
    show_tool_names = Bool(True)

    # Private interface ----------------------------------------------------

    # Cache of tool images (scaled to the appropriate size).
    _image_cache = Instance(ImageCache)

    # ------------------------------------------------------------------------
    # 'object' interface.
    # ------------------------------------------------------------------------

    def __init__(self, *args, **traits):
        """ Creates a new tool bar manager. """

        # Base class contructor.
        super(ToolPaletteManager, self).__init__(*args, **traits)

        # An image cache to make sure that we only load each image used in the
        # tool bar exactly once.
        self._image_cache = ImageCache(self.image_size[0], self.image_size[1])

        return

    # ------------------------------------------------------------------------
    # 'ToolPaletteManager' interface.
    # ------------------------------------------------------------------------

    def create_tool_palette(self, parent, controller=None):
        """ Creates a tool bar. """

        # Create the control.
        tool_palette = ToolPalette(parent)

        # Add all of items in the manager's groups to the tool bar.
        self._add_tools(tool_palette, self.groups)

        self._set_initial_tool_state(tool_palette, self.groups)

        return tool_palette

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _add_tools(self, tool_palette, groups):
        """ Adds tools for all items in a list of groups. """

        previous_non_empty_group = None
        for group in self.groups:
            if len(group.items) > 0:
                # Is a separator required?
                ## FIXME : Does the palette need the notion of a separator?
                ##                 if previous_non_empty_group is not None and group.separator:
                ##                     tool_bar.AddSeparator()
                ##
                ##                 previous_non_empty_group = group

                # Create a tool bar tool for each item in the group.
                for item in group.items:
                    control_id = item.add_to_palette(
                        tool_palette, self._image_cache, self.show_tool_names
                    )
                    item.control_id = control_id

        tool_palette.realize()

    def _set_initial_tool_state(self, tool_palette, groups):
        """ Workaround for the wxPython tool bar bug.

        Without this,  only the first item in a radio group can be selected
         when the tool bar is first realised 8^()

        """

        for group in groups:
            checked = False
            for item in group.items:
                # If the group is a radio group,  set the initial checked state
                # of every tool in it.
                if item.action.style == "radio":
                    tool_palette.toggle_tool(
                        item.control_id, item.action.checked
                    )
                    checked = checked or item.action.checked

                # Every item in a radio group MUST be 'radio' style, so we
                # can just skip to the next group.
                else:
                    break

            # We get here if the group is a radio group.
            else:
                # If none of the actions in the group is specified as 'checked'
                # we will check the first one.
                if not checked and len(group.items) > 0:
                    group.items[0].action.checked = True

        return



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Abstract base class for all action managers. """


from traits.api import Bool, Constant, Event, HasTraits, Instance
from traits.api import List, Property, Str


from pyface.action.action_controller import ActionController
from pyface.action.group import Group


class ActionManager(HasTraits):
    """ Abstract base class for all action managers.

    An action manager contains a list of groups, with each group containing a
    list of items.

    There are currently three concrete sub-classes:

    1) 'MenuBarManager'
    2) 'MenuManager'
    3) 'ToolBarManager'

    """

    # 'ActionManager' interface --------------------------------------------

    #: The Id of the default group.
    DEFAULT_GROUP = Constant("additions")

    #: The action controller (if any) used to control how actions are performed.
    controller = Instance(ActionController)

    #: Is the action manager enabled?
    enabled = Bool(True)

    #: All of the contribution groups in the manager.
    groups = Property(List(Group))

    #: The manager's unique identifier (if it has one).
    id = Str()

    #: Is the action manager visible?
    visible = Bool(True)

    # Events ----

    #: fixme: We probably need more granular events than this!
    changed = Event()

    # Private interface ----------------------------------------------------

    #: All of the contribution groups in the manager.
    _groups = List(Group)

    # ------------------------------------------------------------------------
    # 'object' interface.
    # ------------------------------------------------------------------------

    def __init__(self, *args, **traits):
        """ Creates a new action manager.

        Parameters
        ----------
        args : collection of strings, Group instances, or ActionManagerItem instances
            Positional arguments are interpreted as Items or Groups managed
            by the action manager.

        Notes
        -----

        If a Group is passed as a positional agrument then it is added to the
        manager and any subsequent Items arguments are appended to the Group
        until another Group is encountered.

        If a string is passed, a Group is created with id set to the string.
        """
        # Base class constructor.
        super(ActionManager, self).__init__(**traits)

        # The last group in every manager is the group with Id 'additions'.
        #
        # fixme: The side-effect of this is to ensure that the 'additions'
        # group has been created.  Is the 'additions' group even a good idea?
        group = self._get_default_group()

        # Add all items to the manager.
        for arg in args:
            # We allow a group to be defined by simply specifying a string (its
            # Id).
            if isinstance(arg, str):
                # Create a group with the specified Id.
                arg = Group(id=arg)

            # If the item is a group then add it just before the default group
            # (ie. we always keep the default group as the last group in the
            # manager).
            if isinstance(arg, Group):
                self.insert(-1, arg)
                group = arg

            # Otherwise, the item is an action manager item so add it to the
            # current group.
            else:
                group.append(arg)

    # ------------------------------------------------------------------------
    # 'ActionManager' interface.
    # ------------------------------------------------------------------------

    # Trait properties -----------------------------------------------------

    def _get_groups(self):
        return self._groups[:]

    # Trait change handlers ------------------------------------------------

    def _enabled_changed(self, trait_name, old, new):
        for group in self._groups:
            group.enabled = new

    def _visible_changed(self, trait_name, old, new):
        for group in self._groups:
            group.visible = new

    # Methods -------------------------------------------------------------#

    def append(self, item):
        """ Append an item to the manager.

        Parameters
        ----------
        item : string, Group instance or ActionManagerItem instance
            The item to append.

        Notes
        -----

        If the item is a group, the Group is appended to the manager's list
        of groups.  It the item is a string, then a group is created with
        the string as the ``id`` and the new group is appended to the list
        of groups.  If the item is an ActionManagerItem then the item is
        appended to the manager's default group.
        """
        item = self._prepare_item(item)
        if isinstance(item, Group):
            group = self._groups

        else:
            group = self._get_default_group()

        group.append(item)
        return group

    def destroy(self):
        """ Called when the manager is no longer required.

        By default this method simply calls 'destroy' on all of the manager's
        groups.
        """
        for group in self.groups:
            group.destroy()

    def insert(self, index, item):
        """ Insert an item into the manager at the specified index.

        Parameters
        ----------
        index : int
            The position at which to insert the object
        item : string, Group instance or ActionManagerItem instance
            The item to insert.

        Notes
        -----

        If the item is a group, the Group is inserted into the manager's list
        of groups.  It the item is a string, then a group is created with
        the string as the ``id`` and the new group is inserted into the list
        of groups.  If the item is an ActionManagerItem then the item is
        inserted into the manager's defualt group.
        """
        item = self._prepare_item(item)

        if isinstance(item, Group):
            group = self._groups

        else:
            group = self._get_default_group()

        group.insert(index, item)
        return group

    def find_group(self, id):
        """ Find a group with a specified Id.

        Parameters
        ----------
        id : str
            The id of the group to find.

        Returns
        -------
        group : Group instance
            The group which matches the id, or None if no such group exists.
        """
        for group in self._groups:
            if group.id == id:
                return group
        else:
            return None

    def find_item(self, path):
        """ Find an item using a path.

        Parameters
        ----------
        path : str
            A '/' separated list of contribution Ids.

        Returns
        -------
        item : ActionManagerItem or None
            Returns the matching ActionManagerItem, or None if any component
            of the path is not found.
        """
        components = path.split("/")

        # If there is only one component, then the path is just an Id so look
        # it up in this manager.
        if len(components) > 0:
            item = self._find_item(components[0])
            if len(components) > 1 and item is not None:
                item = item.find_item("/".join(components[1:]))
        else:
            item = None

        return item

    def walk(self, fn):
        """ Walk the manager applying a function at every item.

        The components are walked in pre-order.

        Parameters
        ----------
        fn : callable
            A callable to apply to the tree of groups and items, starting with
            the manager.
        """
        fn(self)

        for group in self._groups:
            self.walk_group(group, fn)

    def walk_group(self, group, fn):
        """ Walk a group applying a function at every item.

        The components are walked in pre-order.

        Parameters
        ----------
        fn : callable
            A callable to apply to the tree of groups and items.
        """
        fn(group)

        for item in group.items:
            if isinstance(item, Group):
                self.walk_group(item, fn)
            else:
                self.walk_item(item, fn)

    def walk_item(self, item, fn):
        """ Walk an item (may be a sub-menu manager remember!).

        The components are walked in pre-order.

        Parameters
        ----------
        fn : callable
            A callable to apply to the tree of items and subgroups.
        """
        if hasattr(item, "groups"):
            item.walk(fn)
        else:
            fn(item)

    # ------------------------------------------------------------------------
    # Private interface.
    # ------------------------------------------------------------------------

    def _get_default_group(self):
        """ Returns the manager's default group.

        This will create this group if it doesn't already exist.

        Returns
        -------
        group : Group instance
            The manager's default group.
        """
        group = self.find_group(self.DEFAULT_GROUP)
        if group is None:
            group = self._prepare_item(self.DEFAULT_GROUP)
            self._groups.append(group)

        return group

    def _prepare_item(self, item):
        """ Prepare an item to be added to this ActionManager.

        Parameters
        ----------
        item : string, Group instance or ActionManagerItem instance
            The item to be added to this ActionManager

        Returns
        -------
        item : Group or ActionManagerItem
            Modified item
        """
        # 1) The item is a 'Group' instance.
        if isinstance(item, Group):
            item.parent = self

        # 2) The item is a string.
        elif isinstance(item, str):
            # Create a group with that Id.
            item = Group(id=item)
            item.parent = self

        return item

    def _find_item(self, id):
        """ Find an item with a spcified Id.

        Parameters
        ----------
        id : str
            The id of the item to be found.

        Returns
        -------
        item : ActionManagerItem or None
            Returns the item with the specified Id, or None if no such item
            exists.
        """
        for group in self.groups:
            item = group.find(id)
            if item is not None:
                return item
        else:
            return None

    # ------------------------------------------------------------------------
    # Debugging interface.
    # ------------------------------------------------------------------------

    def dump(self, indent=""):
        """ Render a manager! """
        print(indent, "Manager", self.id)
        indent += "  "

        for group in self._groups:
            self.render_group(group, indent)

    def render_group(self, group, indent=""):
        """ Render a group! """
        print(indent, "Group", group.id)
        indent += "    "

        for item in group.items:
            if isinstance(item, Group):
                print("Surely, a group cannot contain another group!!!!")
                self.render_group(item, indent)

            else:
                self.render_item(item, indent)

    def render_item(self, item, indent=""):
        """ Render an item! """

        if hasattr(item, "groups"):
            item.dump(indent)

        else:
            print(indent, "Item", item.id)



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import logging

from traits.api import Callable, Enum, Instance, observe, provides

from pyface.qt.QtCore import QAbstractItemModel, QEvent, QObject
from pyface.qt.QtGui import (
    QAbstractItemView, QItemSelection, QItemSelectionModel, QPalette,
    QTreeView
)
from pyface.data_view.i_data_view_widget import (
    IDataViewWidget, MDataViewWidget
)
from pyface.ui.qt4.widget import Widget
from .data_view_item_model import DataViewItemModel

# XXX This file is scaffolding and may need to be rewritten

logger = logging.getLogger(__name__)

qt_selection_types = {
    "row": QAbstractItemView.SelectRows,
    "column": QAbstractItemView.SelectColumns,
    "item": QAbstractItemView.SelectItems,
}
pyface_selection_types = {
    value: key for key, value in qt_selection_types.items()
}
qt_selection_modes = {
    "none": QAbstractItemView.NoSelection,
    "single": QAbstractItemView.SingleSelection,
    "extended": QAbstractItemView.ExtendedSelection,
}
pyface_selection_modes = {
    value: key for key, value in qt_selection_modes.items()
}


class DataViewTreeView(QTreeView):
    """ QTreeView subclass that handles drag and drop via DropHandlers. """

    _widget = None

    def dragEnterEvent(self, event):
        drop_handler = self._get_drop_handler(event)
        if drop_handler is not None:
            event.acceptProposedAction()
        else:
            super().dragEnterEvent(event)

    def dragMoveEvent(self, event):
        drop_handler = self._get_drop_handler(event)
        if drop_handler is not None:
            event.acceptProposedAction()
        else:
            super().dragMoveEvent(event)

    def dragLeaveEvent(self, event):
        drop_handler = self._get_drop_handler(event)
        if drop_handler is not None:
            event.acceptProposedAction()
        else:
            super().dragLeaveEvent(event)

    def dropEvent(self, event):
        drop_handler = self._get_drop_handler(event)
        if drop_handler is not None:
            drop_handler.handle_drop(event, self)
            event.acceptProposedAction()
        else:
            super().dropEvent(event)

    def _get_drop_handler(self, event):
        if self._widget is not None:
            widget = self._widget
            for drop_handler in widget.drop_handlers:
                if drop_handler.can_handle_drop(event, self):
                    return drop_handler
        return None


@provides(IDataViewWidget)
class DataViewWidget(MDataViewWidget, Widget):
    """ The Qt implementation of the DataViewWidget. """

    #: Factory for the underlying Qt control, to facilitate replacement
    control_factory = Callable(DataViewTreeView)

    # IDataViewWidget Interface traits --------------------------------------

    #: What can be selected.  Qt supports additional selection types.
    selection_type = Enum("row", "column", "item")

    #: How selections are modified.  Qt supports turning off selections.
    selection_mode = Enum("extended", "none", "single")

    # IWidget Interface traits ----------------------------------------------

    control = Instance(QAbstractItemView)

    # Private traits --------------------------------------------------------

    #: The QAbstractItemModel instance used by the view.  This will
    #: usually be a DataViewItemModel subclass.
    _item_model = Instance(QAbstractItemModel)

    # ------------------------------------------------------------------------
    # IDataViewWidget Interface
    # ------------------------------------------------------------------------

    def _create_item_model(self):
        """ Create the DataViewItemModel which wraps the data model. """
        self._item_model = DataViewItemModel(
            self.data_model,
            self.selection_type,
            self.exporters,
        )

    def _get_control_header_visible(self):
        """ Method to get the control's header visibility. """
        return not self.control.isHeaderHidden()

    def _set_control_header_visible(self, header_visible):
        """ Method to set the control's header visibility. """
        self.control.setHeaderHidden(not header_visible)

    def _get_control_selection_type(self):
        """ Toolkit specific method to get the selection type. """
        qt_selection_type = self.control.selectionBehavior()
        return pyface_selection_types[qt_selection_type]

    def _set_control_selection_type(self, selection_type):
        """ Toolkit specific method to change the selection type. """
        qt_selection_type = qt_selection_types[selection_type]
        self.control.setSelectionBehavior(qt_selection_type)
        self._item_model.selectionType = selection_type

    def _get_control_selection_mode(self):
        """ Toolkit specific method to get the selection mode. """
        qt_selection_mode = self.control.selectionMode()
        return pyface_selection_modes[qt_selection_mode]

    def _set_control_selection_mode(self, selection_mode):
        """ Toolkit specific method to change the selection mode. """
        qt_selection_mode = qt_selection_modes[selection_mode]
        self.control.setSelectionMode(qt_selection_mode)

    def _get_control_selection(self):
        """ Toolkit specific method to get the selection. """
        indices = self.control.selectedIndexes()
        if self.selection_type == 'row':
            return self._item_model._extract_rows(indices)
        elif self.selection_type == 'column':
            return self._item_model._extract_columns(indices)
        else:
            return self._item_model._extract_indices(indices)

    def _set_control_selection(self, selection):
        """ Toolkit specific method to change the selection. """
        selection_model = self.control.selectionModel()
        select_flags = QItemSelectionModel.Select
        qt_selection = QItemSelection()

        if self.selection_type == 'row':
            select_flags |= QItemSelectionModel.Rows
            for row, column in selection:
                index = self._item_model._to_model_index(row, (0,))
                qt_selection.select(index, index)
        elif self.selection_type == 'column':
            select_flags |= QItemSelectionModel.Columns
            for row, column in selection:
                index = self._item_model._to_model_index(
                    row + (0,), column)
                qt_selection.select(index, index)
        else:
            for row, column in selection:
                index = self._item_model._to_model_index(row, column)
                qt_selection.select(index, index)
        selection_model.clearSelection()
        selection_model.select(qt_selection, select_flags)

    def _observe_control_selection(self, remove=False):
        selection_model = self.control.selectionModel()
        if remove:
            try:
                selection_model.selectionChanged.disconnect(
                  self._update_selection
                )
            except (TypeError, RuntimeError):
                # has already been disconnected
                logger.info("selectionChanged already disconnected")
        else:
            selection_model.selectionChanged.connect(self._update_selection)

    # ------------------------------------------------------------------------
    # IWidget Interface
    # ------------------------------------------------------------------------

    def _create_control(self, parent):
        """ Create the DataViewWidget's toolkit control. """
        self._create_item_model()

        control = self.control_factory(parent)
        control._widget = self
        control.setUniformRowHeights(True)
        control.setAnimated(True)
        control.setDragEnabled(True)
        control.setModel(self._item_model)
        control.setAcceptDrops(True)
        control.setDropIndicatorShown(True)
        return control

    def destroy(self):
        """ Perform any actions required to destroy the control.
        """
        if self.control is not None:
            self.control.setModel(None)

            # ensure that we release references
            self.control._widget = None
            self._item_model = None

        super().destroy()

    # ------------------------------------------------------------------------
    # Private methods
    # ------------------------------------------------------------------------

    # Trait observers

    @observe('data_model', dispatch='ui')
    def _update_item_model(self, event):
        if self._item_model is not None:
            self._item_model.model = event.new

    @observe('exporters.items', dispatch='ui')
    def _update_exporters(self, event):
        if self._item_model is not None:
            self._item_model.exporters = self.exporters



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Reusable MVC-based components for Traits-based applications.
    Part of the TraitsUI project of the Enthought Tool Suite.
"""

try:
    from pyface._version import full_version as __version__
except ImportError:
    __version__ = "not-built"


__requires__ = [
    "importlib-metadata",
    "importlib-resources>=1.1.0",
    "traits>=6.2"
]
__extras_require__ = {
    "wx": ["wxpython>=4", "numpy"],
    "pyqt": ["pyqt>=4.10", "pygments"],
    "pyqt5": ["pyqt5", "pygments"],
    "pyside2": ["pyside2", "shiboken2", "pygments"],
    "test": ["packaging"],
}


# ============================= Test Loader ==================================
def load_tests(loader, standard_tests, pattern):
    """ Custom test loading function that enables test filtering using regex
    exclusion pattern.

    Parameters
    ----------
    loader : unittest.TestLoader
        The instance of test loader
    standard_tests : unittest.TestSuite
        Tests that would be loaded by default from this module (no tests)
    pattern : str
        An inclusion pattern used to match test files (test*.py default)

    Returns
    -------
    filtered_package_tests : unittest.TestSuite
        TestSuite representing all package tests that did not match specified
        exclusion pattern.
    """
    from os import environ
    from os.path import dirname
    from pyface.util.testing import filter_tests, is_traits_version_ge
    from unittest import TestSuite

    # Make sure the right toolkit is up and running before importing tests
    from pyface.toolkit import toolkit_object  # noqa: F401

    this_dir = dirname(__file__)
    package_tests = loader.discover(start_dir=this_dir, pattern=pattern)

    # List of regular expression for filtering test using the test id.
    exclusion_patterns = []

    # Environment variable for skipping more tests.
    # e.g. etstool.py in the source tree root sets this to skip packages for
    # specific toolkit
    additional_exclude = environ.get("EXCLUDE_TESTS", None)
    if additional_exclude is not None:
        exclusion_patterns.append(additional_exclude)

    filtered_package_tests = TestSuite()
    for test_suite in package_tests:
        for exclusion_pattern in exclusion_patterns:
            test_suite = filter_tests(test_suite, exclusion_pattern)
        filtered_package_tests.addTest(test_suite)

    return filtered_package_tests



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import logging

from .checkbox_image_renderer import CheckboxImageRenderer
from .grid_cell_renderer import GridCellRenderer


logger = logging.getLogger(__name__)


class CheckboxRenderer(GridCellRenderer):
    def __init__(self, **traits):

        # base-class constructor
        super(CheckboxRenderer, self).__init__(**traits)

        # initialize the renderer, if it hasn't already been initialized
        if self.renderer is None:
            self.renderer = CheckboxImageRenderer()

    def on_left_click(self, grid, row, col):
        """ Toggles the value. """

        value = grid.model.get_cell_drag_value(row, col)

        try:
            grid.model.set_value(row, col, not value)
        except Exception:
            logger.exception("Can't set cell value")

        return True



# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from unittest import TestCase

from traits.testing.api import UnittestTools
from traits.testing.optional_dependencies import numpy as np, requires_numpy

from pyface.data_view.data_view_errors import DataViewSetError
from pyface.data_view.abstract_value_type import AbstractValueType
from pyface.data_view.value_types.api import (
    FloatValue, IntValue, no_value
)
# This import results in an error without numpy installed
# see enthought/pyface#742
if np is not None:
    from pyface.data_view.data_models.api import ArrayDataModel


@requires_numpy
class TestArrayDataModel(UnittestTools, TestCase):

    def setUp(self):
        super().setUp()
        self.array = np.arange(30.0).reshape(5, 2, 3)
        self.model = ArrayDataModel(data=self.array, value_type=FloatValue())
        self.values_changed_event = None
        self.structure_changed_event = None
        self.model.observe(self.model_values_changed, 'values_changed')
        self.model.observe(self.model_structure_changed, 'structure_changed')

    def tearDown(self):
        self.model.observe(
            self.model_values_changed, 'values_changed', remove=True)
        self.model.observe(
            self.model_structure_changed, 'structure_changed', remove=True)
        self.values_changed_event = None
        self.structure_changed_event = None
        super().tearDown()

    def model_values_changed(self, event):
        self.values_changed_event = event

    def model_structure_changed(self, event):
        self.structure_changed_event = event

    def test_no_data(self):
        model = ArrayDataModel(value_type=FloatValue())
        self.assertEqual(model.data.ndim, 2)
        self.assertEqual(model.data.shape, (0, 0))
        self.assertEqual(model.data.dtype, np.float)
        self.assertEqual(model.get_column_count(), 0)
        self.assertTrue(model.can_have_children(()))
        self.assertEqual(model.get_row_count(()), 0)

    def test_data_1d(self):
        array = np.arange(30.0)
        model = ArrayDataModel(data=array, value_type=FloatValue())
        self.assertEqual(model.data.ndim, 2)
        self.assertEqual(model.data.shape, (30, 1))

    def test_data_list(self):
        data = list(range(30))
        model = ArrayDataModel(data=data, value_type=FloatValue())
        self.assertEqual(model.data.ndim, 2)
        self.assertEqual(model.data.shape, (30, 1))

    def test_set_data_1d(self):
        with self.assertTraitChanges(self.model, 'structure_changed'):
            self.model.data = np.arange(30.0)
        self.assertEqual(self.model.data.ndim, 2)
        self.assertEqual(self.model.data.shape, (30, 1))

    def test_set_data_list(self):
        with self.assertTraitChanges(self.model, 'structure_changed'):
            self.model.data = list(range(30))
        self.assertEqual(self.model.data.ndim, 2)
        self.assertEqual(self.model.data.shape, (30, 1))

    def test_get_column_count(self):
        result = self.model.get_column_count()
        self.assertEqual(result, 3)

    def test_can_have_children(self):
        for row in self.model.iter_rows():
            with self.subTest(row=row):
                result = self.model.can_have_children(row)
                if len(row) <= 1:
                    self.assertEqual(result, True)
                else:
                    self.assertEqual(result, False)

    def test_get_row_count(self):
        for row in self.model.iter_rows():
            with self.subTest(row=row):
                result = self.model.get_row_count(row)
                if len(row) == 0:
                    self.assertEqual(result, 5)
                elif len(row) == 1:
                    self.assertEqual(result, 2)
                else:
                    self.assertEqual(result, 0)

    def test_get_value(self):
        for row, column in self.model.iter_items():
            with self.subTest(row=row, column=column):
                result = self.model.get_value(row, column)
                if len(row) == 0 and len(column) == 0:
                    self.assertIsNone(result)
                elif len(row) == 0:
                    self.assertEqual(result, column[0])
                elif len(column) == 0:
                    self.assertEqual(result, row[-1])
                elif len(row) == 1:
                    self.assertIsNone(result)
                else:
                    self.assertEqual(
                        result,
                        self.array[row[0], row[1], column[0]]
                    )

    def test_set_value(self):
        for row, column in self.model.iter_items():
            with self.subTest(row=row, column=column):
                if len(row) == 0 and len(column) == 0:
                    with self.assertRaises(DataViewSetError):
                        self.model.set_value(row, column, 0)
                elif len(row) == 0:
                    with self.assertRaises(DataViewSetError):
                        self.model.set_value(row, column, column[0] + 1)
                elif len(column) == 0:
                    with self.assertRaises(DataViewSetError):
                        self.model.set_value(row, column, row[-1] + 1)
                elif len(row) == 1:
                    value = 6.0 * row[-1] + 2 * column[0]
                    with self.assertTraitDoesNotChange(
                            self.model, "values_changed"):
                        with self.assertRaises(DataViewSetError):
                            self.model.set_value(row, column, value)
                else:
                    value = 6.0 * row[-1] + 2 * column[0]
                    with self.assertTraitChanges(self.model, "values_changed"):
                        self.model.set_value(row, column, value)
                    self.assertEqual(
                        self.array[row[0], row[1], column[0]],
                        value,
                    )
                    self.assertEqual(
                        self.values_changed_event.new,
                        (row, column, row, column)
                    )

    def test_get_value_type(self):
        for row, column in self.model.iter_items():
            with self.subTest(row=row, column=column):
                result = self.model.get_value_type(row, column)
                if len(row) == 0 and len(column) == 0:
                    self.assertIsInstance(result, AbstractValueType)
                    self.assertIs(result, self.model.label_header_type)
                elif len(row) == 0:
                    self.assertIsInstance(result, AbstractValueType)
                    self.assertIs(result, self.model.column_header_type)
                elif len(column) == 0:
                    self.assertIsInstance(result, AbstractValueType)
                    self.assertIs(result, self.model.row_header_type)
                elif len(row) == 1:
                    self.assertIs(result, no_value)
                else:
                    self.assertIsInstance(result, AbstractValueType)
                    self.assertIs(result, self.model.value_type)

    def test_data_updated(self):
        with self.assertTraitChanges(self.model, "values_changed"):
            self.model.data = 2 * self.array
        self.assertEqual(
            self.values_changed_event.new,
            ((0,), (0,), (4,), (2,))
        )

    def test_data_updated_new_shape(self):
        with self.assertTraitChanges(self.model, "structure_changed"):
            self.model.data = 2 * self.array.T
        self.assertTrue(self.structure_changed_event.new)

    def test_type_updated(self):
        with self.assertTraitChanges(self.model, "values_changed"):
            self.model.value_type = IntValue()
        self.assertEqual(
            self.values_changed_event.new,
            ((0,), (0,), (4,), (2,))
        )

    def test_type_updated_empty(self):
        self.model.data = np.empty((0, 0, 0), dtype='int')
        with self.assertTraitDoesNotChange(self.model, "values_changed"):
            self.model.value_type = IntValue()

    def test_type_attribute_updated(self):
        with self.assertTraitChanges(self.model, "values_changed"):
            self.model.value_type.is_editable = False
        self.assertEqual(
            self.values_changed_event.new,
            ((0,), (0,), (4,), (2,))
        )

    def test_type_attribute_updated_empty(self):
        self.model.data = np.empty((0, 0, 0), dtype='int')
        with self.assertTraitDoesNotChange(self.model, "values_changed"):
            self.model.value_type.is_editable = False

    def test_row_header_type_updated(self):
        with self.assertTraitChanges(self.model, "values_changed"):
            self.model.row_header_type = no_value
        self.assertEqual(
            self.values_changed_event.new,
            ((0,), (), (4,), ())
        )

    def test_row_header_type_updated_empty(self):
        self.model.data = np.empty((0, 4, 2), dtype='int')
        with self.assertTraitDoesNotChange(self.model, "values_changed"):
            self.model.row_header_type = no_value

    def test_row_header_attribute_updated(self):
        with self.assertTraitChanges(self.model, "values_changed"):
            self.model.row_header_type.format = str
        self.assertEqual(
            self.values_changed_event.new,
            ((0,), (), (4,), ())
        )

    def test_row_header_attribute_updated_empty(self):
        self.model.data = np.empty((0, 4, 2), dtype='int')
        with self.assertTraitDoesNotChange(self.model, "values_changed"):
            self.model.row_header_type.format = str

    def test_column_header_type_updated(self):
        with self.assertTraitChanges(self.model, "values_changed"):
            self.model.column_header_type = no_value
        self.assertEqual(
            self.values_changed_event.new,
            ((), (0,), (), (2,))
        )

    def test_row_header_type_updated_empty(self):
        self.model.data = np.empty((2, 4, 0), dtype='int')
        with self.assertTraitDoesNotChange(self.model, "values_changed"):
            self.model.column_header_type = no_value

    def test_column_header_type_attribute_updated(self):
        with self.assertTraitChanges(self.model, "values_changed"):
            self.model.column_header_type.format = str
        self.assertEqual(
            self.values_changed_event.new,
            ((), (0,), (), (2,))
        )

    def test_column_header_attribute_updated_empty(self):
        self.model.data = np.empty((2, 4, 0), dtype='int')
        with self.assertTraitDoesNotChange(self.model, "values_changed"):
            self.model.column_header_type.format = str

    def test_label_header_type_updated(self):
        with self.assertTraitChanges(self.model, "values_changed"):
            self.model.label_header_type = no_value
        self.assertEqual(
            self.values_changed_event.new,
            ((), (), (), ())
        )

    def test_label_header_type_attribute_updated(self):
        with self.assertTraitChanges(self.model, "values_changed"):
            self.model.label_header_type.text = "My Table"
        self.assertEqual(
            self.values_changed_event.new,
            ((), (), (), ())
        )

    def test_is_row_valid(self):
        # valid rows are valid
        for row in self.model.iter_rows():
            with self.subTest(row=row):
                result = self.model.is_row_valid(row)
                self.assertTrue(result)

    def test_is_row_valid_big(self):
        result = self.model.is_row_valid((5,))
        self.assertFalse(result)

    def test_is_row_valid_long(self):
        result = self.model.is_row_valid((1, 1, 1))
        self.assertFalse(result)

    def test_is_column_valid(self):
        # valid columns are valid
        columns = [()] + [(i,) for i in range(3)]
        for column in columns:
            with self.subTest(column=column):
                result = self.model.is_column_valid(column)
                self.assertTrue(result)

    def test_is_column_valid_big(self):
        result = self.model.is_column_valid((3,))
        self.assertFalse(result)

    def test_is_column_valid_long(self):
        result = self.model.is_column_valid((1, 1))
        self.assertFalse(result)

    def test_iter_rows(self):
        result = list(self.model.iter_rows())
        self.assertEqual(
            result,
            [
                (),
                (0,),
                (0, 0),
                (0, 1),
                (1,),
                (1, 0),
                (1, 1),
                (2,),
                (2, 0),
                (2, 1),
                (3,),
                (3, 0),
                (3, 1),
                (4,),
                (4, 0),
                (4, 1),
            ]
        )

    def test_iter_rows_start(self):
        result = list(self.model.iter_rows((2,)))
        self.assertEqual(
            result,
            [(2,), (2, 0), (2, 1)]
        )

    def test_iter_rows_leaf(self):
        result = list(self.model.iter_rows([2, 0]))
        self.assertEqual(result, [(2, 0)])

    def test_iter_items(self):
        result = list(self.model.iter_items())
        self.assertEqual(
            result,
            [
                ((), ()),
                ((), (0,)), ((), (1,)), ((), (2,)),
                ((0,), ()),
                ((0,), (0,)), ((0,), (1,)), ((0,), (2,)),
                ((0, 0), ()),
                ((0, 0), (0,)), ((0, 0), (1,)), ((0, 0), (2,)),
                ((0, 1), ()),
                ((0, 1), (0,)), ((0, 1), (1,)), ((0, 1), (2,)),
                ((1,), ()),
                ((1,), (0,)), ((1,), (1,)), ((1,), (2,)),
                ((1, 0), ()),
                ((1, 0), (0,)), ((1, 0), (1,)), ((1, 0), (2,)),
                ((1, 1), ()),
                ((1, 1), (0,)), ((1, 1), (1,)), ((1, 1), (2,)),
                ((2,), ()),
                ((2,), (0,)), ((2,), (1,)), ((2,), (2,)),
                ((2, 0), ()),
                ((2, 0), (0,)), ((2, 0), (1,)), ((2, 0), (2,)),
                ((2, 1), ()),
                ((2, 1), (0,)), ((2, 1), (1,)), ((2, 1), (2,)),
                ((3,), ()),
                ((3,), (0,)), ((3,), (1,)), ((3,), (2,)),
                ((3, 0), ()),
                ((3, 0), (0,)), ((3, 0), (1,)), ((3, 0), (2,)),
                ((3, 1), ()),
                ((3, 1), (0,)), ((3, 1), (1,)), ((3, 1), (2,)),
                ((4,), ()),
                ((4,), (0,)), ((4,), (1,)), ((4,), (2,)),
                ((4, 0), ()),
                ((4, 0), (0,)), ((4, 0), (1,)), ((4, 0), (2,)),
                ((4, 1), ()),
                ((4, 1), (0,)), ((4, 1), (1,)), ((4, 1), (2,)),
            ]
        )

    def test_iter_items_start(self):
        result = list(self.model.iter_items((2,)))
        self.assertEqual(
            result,
            [
                ((2,), ()),
                ((2,), (0,)), ((2,), (1,)), ((2,), (2,)),
                ((2, 0), ()),
                ((2, 0), (0,)), ((2, 0), (1,)), ((2, 0), (2,)),
                ((2, 1), ()),
                ((2, 1), (0,)), ((2, 1), (1,)), ((2, 1), (2,)),
            ]
        )

    def test_iter_items_leaf(self):
        result = list(self.model.iter_items((2, 0)))
        self.assertEqual(
            result,
            [
                ((2, 0), ()),
                ((2, 0), (0,)), ((2, 0), (1,)), ((2, 0), (2,)),
            ]
        )




# (C) Copyright 2005-2021 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
""" Utilities for handling optional import dependencies."""

import contextlib


@contextlib.contextmanager
def optional_import(dependency_name, msg, logger):
    """ Context manager for capturing ImportError for a particular optional
    dependency. If such an error occurs, it will be silenced and a debug
    message will be logged.

    Parameters
    ----------
    dependency_name : str
        Name of the module that may fail to be imported.
        If matched, the ImportError will be silenced
    msg : str
        Log message to be emitted.
    logger : Logger
        Logger to use for logging messages.
    """
    try:
        yield
    except ImportError as exception:
        if exception.name == dependency_name:
            logger.debug(msg, exc_info=True)
        else:
            raise
